# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from .manager import SubtaskRunnerManagerActor


class SubtaskWorkerService(AbstractService):
    """
    Subtask service on worker.

    Service Configuration
    ---------------------
    {
        "subtask" : {

        }
    }
    """
    async def start(self):
        subtask_config = self._config.get('subtask', dict())
        subtask_processor_cls = subtask_config.get('subtask_processor_cls')
        await mo.create_actor(SubtaskRunnerManagerActor,
                              worker_address=self._address,
                              subtask_processor_cls=subtask_processor_cls,
                              address=self._address,
                              uid=SubtaskRunnerManagerActor.default_uid())

    async def stop(self):
        await mo.destroy_actor(mo.create_actor_ref(
            uid=SubtaskRunnerManagerActor.default_uid(), address=self._address))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _install():
    import pandas as pd

    from ..base.accessor import CachedAccessor
    from ..core import DATAFRAME_TYPE, SERIES_TYPE
    from .core import PlotAccessor

    for t in DATAFRAME_TYPE + SERIES_TYPE:
        t.plot = CachedAccessor('plot', PlotAccessor)

    for method in dir(pd.DataFrame.plot):
        if not method.startswith('_'):
            PlotAccessor._register(method)

    PlotAccessor.__doc__ = pd.DataFrame.plot.__doc__.replace('pd.', 'md.')


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..utils import make_import_error_func
from .core import xgboost, XGBScikitLearnBase


XGBRegressor = make_import_error_func('xgboost')
if xgboost:
    from .core import wrap_evaluation_matrices
    from .train import train
    from .predict import predict

    class XGBRegressor(XGBScikitLearnBase):
        """
        Implementation of the scikit-learn API for XGBoost regressor.
        """

        def fit(self, X, y, sample_weight=None, base_margin=None,
                eval_set=None, sample_weight_eval_set=None, base_margin_eval_set=None, **kw):
            session = kw.pop('session', None)
            run_kwargs = kw.pop('run_kwargs', dict())
            if kw:
                raise TypeError(f"fit got an unexpected keyword argument '{next(iter(kw))}'")

            dtrain, evals = wrap_evaluation_matrices(
                None, X, y, sample_weight, base_margin, eval_set,
                sample_weight_eval_set, base_margin_eval_set)
            params = self.get_xgb_params()
            self.evals_result_ = dict()
            result = train(params, dtrain, num_boost_round=self.get_num_boosting_rounds(),
                           evals=evals, evals_result=self.evals_result_,
                           session=session, run_kwargs=run_kwargs)
            self._Booster = result
            return self

        def predict(self, data, **kw):
            session = kw.pop('session', None)
            run_kwargs = kw.pop('run_kwargs', None)
            return predict(self.get_booster(), data,
                           session=session, run_kwargs=run_kwargs, **kw)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...utils import classproperty
from .core import DataFrameBinopUfunc
from .docstring import bin_compare_doc


class DataFrameNotEqual(DataFrameBinopUfunc):
    _op_type_ = OperandDef.NE

    _func_name = 'ne'
    _rfunc_name = 'ne'

    @classproperty
    def _operator(self):
        return lambda lhs, rhs: lhs.ne(rhs)

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorNotEqual
        return TensorNotEqual


_ne_example = """
>>> a.ne(b, fill_value=0).execute()
a    False
b     True
c     True
d     True
e     True
dtype: bool
"""


@bin_compare_doc('Not equal to', equiv='!=', series_example=_ne_example)
def ne(df, other, axis='columns', level=None):
    op = DataFrameNotEqual(axis=axis, level=level, lhs=df, rhs=other)
    return op(df, other)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...core.operand.base import SchedulingHint
from ...serialization.serializables import StringField
from ...tensor.datastore.to_vineyard import resolve_vineyard_socket
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index

try:
    import vineyard
    from vineyard.data.dataframe import make_global_dataframe
    from vineyard.data.utils import to_json
except ImportError:
    vineyard = None


class DataFrameToVineyardChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_CHUNK

    # vineyard ipc socket
    vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, dtypes=None, **kw):
        super().__init__(vineyard_socket=vineyard_socket, _dtypes=dtypes,
                         _output_types=[OutputType.dataframe], **kw)

    def __call__(self, df):
        return self.new_dataframe([df], shape=(0, 0), dtypes=df.dtypes,
                                  index_value=df.index_value,
                                  columns_value=df.columns_value)

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        dtypes = pd.Series([np.dtype('O')], index=pd.Index([0]))
        merge_op = DataFrameToVinyardStoreMeta(
                vineyard_socket=op.vineyard_socket,
                chunk_shape=op.inputs[0].chunk_shape,
                shape=(1, 1), dtypes=dtypes)
        return merge_op.new_chunks(out_chunks, shape=(1, 1), dtypes=dtypes,
                                   index=(0, 0))

    @classmethod
    def tile(cls, op):
        out_chunks = []
        scheduling_hint = SchedulingHint(fuseable=False)
        dtypes = pd.Series([np.dtype('O')], index=pd.Index([0]))
        for idx, chunk in enumerate(op.inputs[0].chunks):
            chunk_op = op.copy().reset_key()
            chunk_op.scheduling_hint = scheduling_hint
            out_chunk = chunk_op.new_chunk([chunk], shape=(1, 1), dtypes=dtypes,
                                           index_value=chunk.index_value,
                                           columns_value=chunk.columns_value,
                                           index=(idx, 0))
            out_chunks.append(out_chunk)
        out_chunks = cls._process_out_chunks(op, out_chunks)

        in_df = op.inputs[0]
        new_op = op.copy().reset_key()
        return new_op.new_dataframes(op.inputs, shape=(len(out_chunks), 1),
                                     dtypes=dtypes,
                                     index_value=in_df.index_value,
                                     columns_value=in_df.columns_value,
                                     chunks=out_chunks,
                                     nsplits=((np.prod(op.inputs[0].chunk_shape),),))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')

        socket, needs_put = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        # some op might be fused and executed twice on different workers
        if not needs_put:
            # might be fused
            try:  # pragma: no cover
                meta = ctx.get_chunks_meta([op.inputs[0].key])[0]
                df_id = vineyard.ObjectID(meta['object_ref'])
                if not client.exists(df_id):
                    needs_put = True
            except KeyError:
                needs_put = True
        if needs_put:
            df_id = client.put(ctx[op.inputs[0].key],
                               partition_index=op.inputs[0].index)
        else:  # pragma: no cover
            meta = client.get_meta(df_id)
            new_meta = vineyard.ObjectMeta()
            for k, v in meta.items():
                if k not in ['id', 'signature', 'instance_id']:
                    if isinstance(v, vineyard.ObjectMeta):
                        new_meta.add_member(k, v)
                    else:
                        new_meta[k] = v
            new_meta['partition_index_'] = to_json(op.inputs[0].index)
            df_id = client.create_metadata(new_meta).id

        client.persist(df_id)
        ctx[op.outputs[0].key] = pd.DataFrame({0: [df_id]})


class DataFrameToVinyardStoreMeta(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_META

    # vineyard ipc socket
    vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, dtypes=None, **kw):
        super().__init__(vineyard_socket=vineyard_socket, dtypes=dtypes,
                         _output_types=[OutputType.dataframe], **kw)

    @classmethod
    def tile(cls, op):
        dtypes = pd.Series([np.dtype('O')], index=pd.Index([0]))
        chunk_op = op.copy().reset_key()
        out_chunk = chunk_op.new_chunk(op.inputs[0].chunks, shape=(1, 1), dtypes=dtypes,
                                       index_value=parse_index(pd.Index([-1])),
                                       columns_value=parse_index(pd.Index([0])),
                                       index=(0, 0))
        new_op = op.copy().reset_key()
        return new_op.new_dataframes(op.inputs, shape=(1, 1),
                                     dtypes=dtypes,
                                     index_value=parse_index(pd.Index([0])),
                                     columns_value=parse_index(pd.Index([0])),
                                     chunks=[out_chunk],
                                     nsplits=((1,), (1,)))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')

        socket, _ = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        # # store the result object id to execution context
        chunks = [ctx[chunk.key][0][0] for chunk in op.inputs]
        ctx[op.outputs[0].key] = pd.DataFrame(
                {0: [make_global_dataframe(client, chunks).id]})


def to_vineyard(df, vineyard_socket=None):
    op = DataFrameToVineyardChunk(vineyard_socket=vineyard_socket)
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ...services import NodeRole
from ...utils import get_next_port
from .cmdline import OscarCommandRunner
from .local import start_supervisor, stop_supervisor
from .pool import create_supervisor_actor_pool


class SupervisorCommandRunner(OscarCommandRunner):
    command_description = 'Mars Supervisor'
    node_role = NodeRole.SUPERVISOR

    def __init__(self):
        super().__init__()
        self._endpoint_file_name = None

    def config_args(self, parser):
        super().config_args(parser)
        parser.add_argument('-w', '--web-port', help='web port of the service')

    def parse_args(self, parser, argv, environ=None):
        args = super().parse_args(parser, argv, environ=environ)

        if args.endpoint is None:
            args.endpoint = f'{args.host}:{get_next_port()}'
        self._endpoint_file_name = self._write_supervisor_endpoint_file(args)

        args.supervisors = f'{args.supervisors},{args.endpoint}'.strip(',')

        web_config = self.config.get('web', {})
        if args.web_port is not None:
            web_config['port'] = int(args.web_port)
        self.config['web'] = web_config

        return args

    async def create_actor_pool(self):
        return await create_supervisor_actor_pool(
            self.args.endpoint, n_process=0, ports=self.ports,
            modules=self.args.load_modules,
            logging_conf=self.logging_conf,
            subprocess_start_method='forkserver' if os.name == 'nt' else 'spawn'
        )

    async def start_services(self):
        return await start_supervisor(
            self.pool.external_address, self.args.supervisors,
            self.args.load_modules, self.config)

    async def stop_services(self):
        if self._endpoint_file_name is not None:  # pragma: no branch
            try:
                os.unlink(self._endpoint_file_name)
            except OSError:  # pragma: no cover
                pass
        return await stop_supervisor(self.pool.external_address, self.config)


main = SupervisorCommandRunner()

if __name__ == '__main__':  # pragma: no branch
    main()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import multiprocessing
from typing import List, Dict

from ..config import ActorPoolConfig
from ..communication import gen_local_address, get_server_type, DummyServer
from ..mars.pool import MainActorPool, SubActorPool, SubpoolStatus
from ..pool import ActorPoolType


class TestMainActorPool(MainActorPool):
    @classmethod
    def get_external_addresses(
            cls, address: str, n_process: int = None, ports: List[int] = None):
        if '://' in address:
            address = address.split('://', 1)[1]
        return super().get_external_addresses(address, n_process=n_process,
                                              ports=ports)

    @classmethod
    def gen_internal_address(cls, process_index: int, external_address: str = None) -> str:
        return f'dummy://{process_index}'

    @classmethod
    async def start_sub_pool(
            cls,
            actor_pool_config: ActorPoolConfig,
            process_index: int,
            start_method: str = None):
        status_queue = multiprocessing.Queue()
        return asyncio.create_task(
            cls._create_sub_pool(actor_pool_config, process_index, status_queue))

    @classmethod
    async def wait_sub_pools_ready(cls,
                                   create_pool_tasks: List[asyncio.Task]):
        return [await t for t in create_pool_tasks]

    @classmethod
    async def _create_sub_pool(
            cls,
            actor_config: ActorPoolConfig,
            process_index: int,
            status_queue: multiprocessing.Queue):
        pool = await TestSubActorPool.create({
            'actor_pool_config': actor_config,
            'process_index': process_index
        })
        await pool.start()
        status_queue.put(SubpoolStatus(0))
        await pool.join()

    async def kill_sub_pool(self, process: multiprocessing.Process,
                            force: bool = False):
        process.cancel()

    async def is_sub_pool_alive(self, process: multiprocessing.Process):
        return not process.cancelled()


class TestSubActorPool(SubActorPool):
    @classmethod
    async def create(cls, config: Dict) -> ActorPoolType:
        kw = dict()
        cls._parse_config(config, kw)
        process_index: int = kw['process_index']
        actor_pool_config = kw['config']
        external_addresses = \
            actor_pool_config.get_pool_config(process_index)['external_address']

        def handle_channel(channel):
            return pool.on_new_channel(channel)

        # create servers
        create_server_tasks = []
        for addr in set(external_addresses + [gen_local_address(process_index)]):
            server_type = get_server_type(addr)
            task = asyncio.create_task(
                server_type.create(dict(address=addr,
                                        handle_channel=handle_channel)))
            create_server_tasks.append(task)
        await asyncio.gather(*create_server_tasks)
        kw['servers'] = [f.result() for f in create_server_tasks]

        # create pool
        pool = cls(**kw)
        return pool

    async def stop(self):
        # do not close dummy server
        self._servers = [s for s in self._servers[:-1]
                         if not isinstance(s, DummyServer)]
        await super().stop()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union, Type

import numpy as np
import pandas as pd

from ...core import OBJECT_TYPE, OBJECT_CHUNK_TYPE
from ...dataframe.core import DtypesValue, IndexValue, \
    DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE, DATAFRAME_GROUPBY_TYPE, \
    SERIES_GROUPBY_TYPE, \
    DATAFRAME_CHUNK_TYPE, SERIES_CHUNK_TYPE, INDEX_CHUNK_TYPE, \
    DATAFRAME_GROUPBY_CHUNK_TYPE, SERIES_GROUPBY_CHUNK_TYPE, \
    CATEGORICAL_TYPE, CATEGORICAL_CHUNK_TYPE
from ...tensor.core import TensorOrder, TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ...typing import BandType
from ...utils import dataslots

PandasDtypeType = Union[np.dtype, pd.api.extensions.ExtensionDtype]

_type_to_meta_class = dict()


def _register_type(object_types: Tuple):
    def _call(meta):
        for obj_type in object_types:
            _type_to_meta_class[obj_type] = meta
        return meta
    return _call


def get_meta_type(object_type: Type) -> Type["_CommonMeta"]:
    try:
        return _type_to_meta_class[object_type]
    except KeyError:
        for m_type in object_type.__mro__:
            try:
                return _type_to_meta_class[m_type]
            except KeyError:
                continue
        raise


@dataslots
@dataclass
class _CommonMeta:
    """
    Class for common meta, for both tileable and chunk, or DataFrame, tensor etc.
    """
    object_id: str
    name: Any = None
    memory_size: int = None # size in memory
    store_size: int = None  # size that stored in storage
    extra: Dict = None


@dataslots
@dataclass
class _TileableMeta(_CommonMeta):
    nsplits: Tuple[Tuple[int]] = None


@_register_type(TENSOR_TYPE)
@dataslots
@dataclass
class TensorMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: np.dtype = None
    order: TensorOrder = None


@_register_type(DATAFRAME_TYPE)
@dataslots
@dataclass
class DataFrameMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None


@_register_type(SERIES_TYPE)
@dataslots
@dataclass
class SeriesMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@_register_type(INDEX_TYPE)
@dataslots
@dataclass
class IndexMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@_register_type(DATAFRAME_GROUPBY_TYPE)
@dataslots
@dataclass
class DataFrameGroupByMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None
    selection: List = None


@_register_type(SERIES_GROUPBY_TYPE)
@dataslots
@dataclass
class SeriesGroupByMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None
    selection: List = None


@_register_type(CATEGORICAL_TYPE)
@dataslots
@dataclass
class CategoricalMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    categories_value: IndexValue = None


@_register_type(OBJECT_TYPE)
@dataslots
@dataclass
class ObjectMeta(_TileableMeta):
    pass


@_register_type(OBJECT_CHUNK_TYPE)
@dataslots
@dataclass
class _ChunkMeta(_CommonMeta):
    index: Tuple[int] = None
    bands: List[BandType] = None
    # needed by ray ownership to keep object alive when worker died.
    object_ref: Any = None


@_register_type(TENSOR_CHUNK_TYPE)
@dataslots
@dataclass
class TensorChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: np.dtype = None
    order: TensorOrder = None


@_register_type(DATAFRAME_CHUNK_TYPE)
@dataslots
@dataclass
class DataFrameChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None


@_register_type(SERIES_CHUNK_TYPE)
@dataslots
@dataclass
class SeriesChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@_register_type(INDEX_CHUNK_TYPE)
@dataslots
@dataclass
class IndexChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@_register_type(DATAFRAME_GROUPBY_CHUNK_TYPE)
@dataslots
@dataclass
class DataFrameGroupByChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None
    selection: List = None


@_register_type(SERIES_GROUPBY_CHUNK_TYPE)
@dataslots
@dataclass
class SeriesGroupByChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None
    selection: List = None


@_register_type(CATEGORICAL_CHUNK_TYPE)
@dataslots
@dataclass
class CategoricalChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    categories_value: IndexValue = None


@_register_type(OBJECT_CHUNK_TYPE)
@dataslots
@dataclass
class ObjectChunkMeta(_ChunkMeta):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
from typing import Dict

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import FieldTypes, KeyField, \
    StringField, BytesField, TupleField
from ...lib.filesystem import get_fs, FSMap
from ...utils import has_unknown_shape
from .core import TensorDataStore


class ZarrOptions(object):
    def __init__(self, options: Dict):
        self._options = options

    def todict(self):
        return self._options

    @staticmethod
    def _stringfy(v):
        return pickle.dumps(v) if not isinstance(v, str) else v

    def __mars_tokenize__(self):
        return list(self._options.keys()), \
               list(self._stringfy(v) for v in self._options.values())

    def __getstate__(self):
        return self._options

    def __setstate__(self, state):
        self._options = state


class TensorToZarrDataStore(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_ZARR

    _input = KeyField('input')
    _path = StringField('path')
    _group = StringField('group')
    _dataset = StringField('dataset')
    _zarr_options = BytesField('zarr_options', on_serialize=pickle.dumps,
                               on_deserialize=pickle.loads)
    _axis_offsets = TupleField('axis_offsets', FieldTypes.int32)

    def __init__(self, path=None, group=None, dataset=None, zarr_options=None,
                 axis_offsets=None, **kw):
        super().__init__(_path=path, _group=group, _dataset=dataset,
                         _zarr_options=zarr_options, _axis_offsets=axis_offsets, **kw)

    @property
    def path(self):
        return self._path

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def zarr_options(self):
        return self._zarr_options

    @property
    def axis_offsets(self):
        return self._axis_offsets

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        import zarr

        if has_unknown_shape(*op.inputs):
            yield
        in_tensor = op.input

        # create dataset
        fs = get_fs(op.path, None)
        path = op.path
        if op.group is not None:
            path += '/' + op.group
        fs_map = FSMap(path, fs)
        zarr.open(fs_map, 'w', path=op.dataset,
                  dtype=in_tensor.dtype, shape=in_tensor.shape,
                  chunks=tuple(max(ns) for ns in in_tensor.nsplits),
                  **op.zarr_options.todict())

        cum_nsplits = [[0] + np.cumsum(ns).tolist() for ns in in_tensor.nsplits]
        out_chunks = []
        for chunk in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._axis_offsets = \
                tuple(cs[i] for i, cs in zip(chunk.index, cum_nsplits))
            out_chunks.append(chunk_op.new_chunk([chunk], shape=(0,) * chunk.ndim,
                                                 index=chunk.index))

        new_op = op.copy()
        out = op.outputs[0]
        nsplits = tuple((0,) * len(ns) for ns in in_tensor.nsplits)
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  nsplits=nsplits, chunks=out_chunks)

    @classmethod
    def execute(cls, ctx, op):
        import zarr

        fs = get_fs(op.path, None)
        fs_map = FSMap(op.path, fs)

        group = zarr.Group(store=fs_map, path=op.group)
        array = group[op.dataset]

        to_store = ctx[op.inputs[0].key]
        axis_offsets = op.axis_offsets
        shape = to_store.shape

        array[tuple(slice(offset, offset + size)
              for offset, size
              in zip(axis_offsets, shape))] = to_store

        ctx[op.outputs[0].key] = np.empty((0,) * to_store.ndim,
                                          dtype=to_store.dtype)


def tozarr(path, x, group=None, dataset=None, **zarr_options):
    import zarr

    if isinstance(path, zarr.Array):
        arr = path
        if isinstance(arr.store, FSMap):
            root = arr.store.root
            path, dataset = root.rsplit('/', 1)
        else:
            path = arr.store.path
            if '/' in arr.path and group is None:
                group = arr.path.rsplit('/', 1)[0]
            dataset = arr.basename
            if not dataset:
                path, dataset = path.rsplit('/', 1)
        for attr in ['compressor', 'filters']:
            if getattr(arr, attr):
                zarr_options[attr] = getattr(arr, attr)
    elif isinstance(path, str):
        if dataset is None:
            path, dataset = path.rsplit('/', 1)
    else:
        raise TypeError('`path` passed has wrong type, '
                        'expect str, or zarr.Array'
                        f'got {type(path)}')

    op = TensorToZarrDataStore(path=path, group=group, dataset=dataset,
                               zarr_options=ZarrOptions(zarr_options))
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from ....lib.aio import alru_cache
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from .core import AbstractLifecycleAPI


class LifecycleWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = '/api/session/(?P<session_id>[^/]+)/lifecycle'

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_lifecycle_api(self, session_id: str):
        from .oscar import LifecycleAPI
        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys([session_id])
        return await LifecycleAPI.create(session_id, address)

    @web_api('', method='post', arg_filter={'action': 'decref_tileables'})
    async def decref_tileables(self, session_id: str):
        tileable_keys = self.get_argument('tileable_keys').split(',')

        oscar_api = await self._get_oscar_lifecycle_api(session_id)
        await oscar_api.decref_tileables(tileable_keys)

    @web_api('', method='get', arg_filter={'action': 'get_all_chunk_ref_counts'})
    async def get_all_chunk_ref_counts(self, session_id: str):
        oscar_api = await self._get_oscar_lifecycle_api(session_id)
        res = await oscar_api.get_all_chunk_ref_counts()
        self.write(serialize_serializable(res))


web_handlers = {
    LifecycleWebAPIHandler.get_root_pattern(): LifecycleWebAPIHandler
}


class WebLifecycleAPI(AbstractLifecycleAPI, MarsWebAPIClientMixin):
    def __init__(self, session_id: str, address: str):
        self._session_id = session_id
        self._address = address.rstrip('/')

    async def decref_tileables(self, tileable_keys: List[str]):
        path = f'{self._address}/api/session/{self._session_id}/lifecycle'
        params = dict(action='decref_tileables')
        await self._request_url(
            path=path, method='POST', params=params,
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            data='tileable_keys=' + ','.join(tileable_keys)
        )

    async def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        params = dict(action='get_all_chunk_ref_counts')
        path = f'{self._address}/api/session/{self._session_id}/lifecycle'
        res = await self._request_url('GET', path, params=params)
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from ..core import StorageManagerActor


class StorageWorkerService(AbstractService):
    """
    Storage service on worker

    Service Configuration
    ---------------------
    {
        "storage": {
            "backends": ["plasma"],
            "<storage backend name>"： "<setup params>",
        }
    }
    """
    async def start(self):
        storage_configs = self._config['storage']
        backends = storage_configs.get('backends')
        options = storage_configs.get('default_config', dict())
        transfer_block_size = options.get('transfer_block_size', None)
        backend_config = {backend: storage_configs.get(backend, dict())
                          for backend in backends}

        await mo.create_actor(StorageManagerActor,
                              backend_config,
                              transfer_block_size,
                              uid=StorageManagerActor.default_uid(),
                              address=self._address)

    async def stop(self):
        await mo.destroy_actor(mo.create_actor_ref(
            address=self._address, uid=StorageManagerActor.default_uid()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import inject_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorSignbit(TensorUnaryOp):
    _op_type_ = OperandDef.SIGNBIT
    _func_name = 'signbit'


@inject_dtype(np.bool_)
def signbit(x, out=None, where=None, **kwargs):
    """
    Returns element-wise True where signbit is set (less than zero).

    Parameters
    ----------
    x : array_like
        The input value(s).
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    result : Tensor of bool
        Output tensor, or reference to `out` if that was supplied.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.signbit(-1.2).execute()
    True
    >>> mt.signbit(mt.array([1, -2.3, 2.1])).execute()
    array([False,  True, False])
    """
    op = TensorSignbit(**kwargs)
    return op(x, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from collections import deque
from typing import Dict, List, Tuple, Type, Union

from ....core import ChunkGraph, ChunkType, enter_mode
from ....core.operand import Fetch, Fuse, VirtualOperand
from ....typing import BandType
from ....utils import build_fetch
from ...subtask import SubtaskGraph, Subtask
from ..core import Task, new_task_id
from .assigner import AbstractGraphAssigner, GraphAssigner
from .fusion import Fusion

logger = logging.getLogger(__name__)


class GraphAnalyzer:
    def __init__(self,
                 chunk_graph: ChunkGraph,
                 band_slots: Dict[BandType, int],
                 task: Task,
                 graph_assigner_cls: Type[AbstractGraphAssigner] = None):
        self._chunk_graph = chunk_graph
        self._band_slots = band_slots
        self._task = task
        self._fuse_enabled = task.fuse_enabled
        self._extra_config = task.extra_config
        if graph_assigner_cls is None:
            graph_assigner_cls = GraphAssigner
        self._graph_assigner_cls = graph_assigner_cls
        self._chunk_to_copied = dict()

    @classmethod
    def _iter_start_ops(cls, chunk_graph: ChunkGraph):
        visited = set()
        op_keys = set()
        start_chunks = deque(chunk_graph.iter_indep())
        stack = deque([start_chunks.popleft()])

        while stack:
            chunk = stack.popleft()
            if chunk not in visited:
                inp_chunks = chunk_graph.predecessors(chunk)
                if not inp_chunks or \
                        all(inp_chunk in visited for inp_chunk in inp_chunks):
                    if len(inp_chunks) == 0:
                        op_key = chunk.op.key
                        if op_key not in op_keys:
                            op_keys.add(op_key)
                            yield chunk.op
                    visited.add(chunk)
                    stack.extend(c for c in chunk_graph[chunk]
                                 if c not in visited)
                else:
                    stack.appendleft(chunk)
                    stack.extendleft(
                        reversed([c for c in chunk_graph.predecessors(chunk)
                                  if c not in visited]))
            if not stack and start_chunks:
                stack.appendleft(start_chunks.popleft())

    def _fuse(self, chunk_to_bands: Dict[ChunkType, BandType]) \
            -> Dict[ChunkType, BandType]:
        fusion = Fusion(self._chunk_graph)
        to_fuse_nodes_list, fused_nodes = fusion.fuse()
        assert len(to_fuse_nodes_list) == len(fused_nodes)
        fuse_to_fused = dict()
        for to_fuse_nodes, fused_node in zip(to_fuse_nodes_list, fused_nodes):
            priority = None
            for to_fuse_node in to_fuse_nodes:
                if to_fuse_node.op.priority:
                    assert priority is None or priority == to_fuse_node.op.priority
                    priority = to_fuse_node.op.priority
                fuse_to_fused[to_fuse_node] = fused_node
            if priority:
                fused_node.op.priority = priority
        # modify chunk_to_bands if some chunk fused
        new_chunk_to_bands = dict()
        for chunk, band in chunk_to_bands.items():
            if chunk in fuse_to_fused:
                new_chunk_to_bands[fuse_to_fused[chunk]] = band
            else:
                new_chunk_to_bands[chunk] = band
        return new_chunk_to_bands

    @classmethod
    def _gen_input_chunks(cls,
                          inp_chunks: List[ChunkType],
                          chunk_to_fetch_chunk: Dict[ChunkType, ChunkType]) \
            -> List[ChunkType]:
        # gen fetch chunks for input chunks
        inp_fetch_chunks = []
        for inp_chunk in inp_chunks:
            if inp_chunk in chunk_to_fetch_chunk:
                inp_fetch_chunks.append(chunk_to_fetch_chunk[inp_chunk])
            elif isinstance(inp_chunk.op, Fetch):
                chunk_to_fetch_chunk[inp_chunk] = inp_chunk
                inp_fetch_chunks.append(inp_chunk)
            else:
                fetch_chunk = build_fetch(inp_chunk).data
                chunk_to_fetch_chunk[inp_chunk] = fetch_chunk
                inp_fetch_chunks.append(fetch_chunk)

        return inp_fetch_chunks

    def _build_subtask_chunk_graph(self,
                                   chunk: ChunkType,
                                   chunk_to_fetch_chunk: Dict[ChunkType, ChunkType]) \
            -> ChunkGraph:
        virtual = isinstance(chunk.op, VirtualOperand)
        inp_chunks = chunk.inputs
        assert all(inp_chunk in self._chunk_graph
                   for inp_chunk in inp_chunks)
        inp_fetch_chunks = self._gen_input_chunks(inp_chunks, chunk_to_fetch_chunk)

        # gen chunk graph for each subtask
        result_chunks = []
        subtask_chunk_graph = ChunkGraph(result_chunks)
        outs = chunk.op.outputs
        out_params = [out.params for out in outs]
        copied_op = chunk.op.copy()
        copied_op._key = chunk.op.key
        copied_chunks = [c.data for c in
                         copied_op.new_chunks(inp_fetch_chunks, kws=out_params)]
        for out, copied_chunk in zip(outs, copied_chunks):
            copied_chunk._key = out.key
            result_chunks.append(copied_chunk)
            subtask_chunk_graph.add_node(copied_chunk)
            if not virtual:
                # skip adding fetch chunk to chunk graph when op is virtual operand
                for inp_fetch_chunk in inp_fetch_chunks:
                    subtask_chunk_graph.add_node(inp_fetch_chunk)
                    subtask_chunk_graph.add_edge(inp_fetch_chunk, copied_chunk)
            self._chunk_to_copied[out] = copied_chunk
        return subtask_chunk_graph

    def _build_fuse_subtask_chunk_graph(self,
                                        chunk: ChunkType,
                                        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType]) \
            -> ChunkGraph:
        # gen chunk graph for each subtask
        result_chunks = []
        subtask_chunk_graph = ChunkGraph(result_chunks)
        fused_chunks = chunk.composed
        fuse_to_copied = dict()
        for i, fuse_chunk in enumerate(fused_chunks):
            copied_op = fuse_chunk.op.copy()
            copied_op._key = fuse_chunk.op.key
            if i == 0:
                # the first chunk
                inp_chunks = chunk.inputs
                inp_fetch_chunks = self._gen_input_chunks(
                    inp_chunks, chunk_to_fetch_chunk)
                copied_fuse_chunk = copied_op.new_chunk(
                    inp_fetch_chunks, kws=[fuse_chunk.params.copy()]).data
                copied_fuse_chunk._key = fuse_chunk.key
                subtask_chunk_graph.add_node(copied_fuse_chunk)
                for inp_fetch_chunk in inp_fetch_chunks:
                    subtask_chunk_graph.add_node(inp_fetch_chunk)
                    subtask_chunk_graph.add_edge(
                        inp_fetch_chunk, copied_fuse_chunk)
            else:
                inp_chunk = fuse_to_copied[fused_chunks[i - 1]]
                copied_fuse_chunk = copied_op.new_chunks(
                    [inp_chunk] * len(fuse_chunk.inputs),
                    kws=[fuse_chunk.params.copy()])[0].data
                copied_fuse_chunk._key = fuse_chunk.key
                subtask_chunk_graph.add_node(copied_fuse_chunk)
                subtask_chunk_graph.add_edge(inp_chunk, copied_fuse_chunk)
                if i == len(fused_chunks) - 1:
                    # the last chunk
                    result_chunks.append(copied_fuse_chunk)
            fuse_to_copied[fuse_chunk] = copied_fuse_chunk
        self._chunk_to_copied[chunk.chunk] = self._chunk_to_copied[chunk] = \
            fuse_to_copied[chunk.chunk]
        return subtask_chunk_graph

    def _gen_subtask(self,
                     chunk: ChunkType,
                     chunk_to_priorities: Dict[ChunkType, Tuple[int, ...]],
                     chunk_to_bands: Dict[ChunkType, BandType],
                     chunk_to_fetch_chunk: Dict[ChunkType, ChunkType]) -> Subtask:
        virtual = isinstance(chunk.op, VirtualOperand)
        inp_chunks = [inp_chunk for inp_chunk in self._chunk_graph.predecessors(chunk)
                      if not isinstance(inp_chunk.op, Fetch)]
        # calculate priority
        if inp_chunks:
            depth = max(chunk_to_priorities[inp_chunk][0]
                        for inp_chunk in inp_chunks) + 1
        else:
            depth = 0
        priority = (depth, chunk.op.priority or 0)
        for out in chunk.op.outputs:
            chunk_to_priorities[out] = priority

        band = chunk_to_bands.get(chunk)
        if not isinstance(chunk.op, Fuse):
            subtask_chunk_graph = self._build_subtask_chunk_graph(
                chunk, chunk_to_fetch_chunk)
        else:
            subtask_chunk_graph = self._build_fuse_subtask_chunk_graph(
                chunk, chunk_to_fetch_chunk)

        subtask = Subtask(
            subtask_id=new_task_id(),
            session_id=self._task.session_id,
            task_id=self._task.task_id,
            chunk_graph=subtask_chunk_graph,
            expect_bands=[band] if band is not None else None,
            virtual=virtual,
            priority=priority,
            retryable=all(getattr(chunk.op, 'retryable', True) for chunk in subtask_chunk_graph),
            extra_config=self._extra_config)
        return subtask

    @staticmethod
    def _to_band(band_or_worker: Union[BandType, str]) -> BandType:
        if isinstance(band_or_worker, tuple) and len(band_or_worker) == 2:
            # band already
            return band_or_worker
        else:
            return band_or_worker, 'numa-0'

    @enter_mode(build=True)
    def gen_subtask_graph(self) -> SubtaskGraph:
        """
        Analyze chunk graph and generate subtask graph.

        Returns
        -------
        subtask_graph: SubtaskGraph
            Subtask graph.
        """
        fetch_removed_chunk_graph = self._chunk_graph.copy()
        reassign_worker_ops = []
        for chunk in self._chunk_graph:
            if isinstance(chunk.op, Fetch):
                fetch_removed_chunk_graph.remove_node(chunk)
            elif chunk.op.reassign_worker:
                reassign_worker_ops.append(chunk.op)

        start_ops = list(self._iter_start_ops(fetch_removed_chunk_graph)) \
            if len(fetch_removed_chunk_graph) > 0 else []

        # assign start chunks
        to_assign_ops = start_ops + reassign_worker_ops
        assigner = self._graph_assigner_cls(
            fetch_removed_chunk_graph, to_assign_ops, self._band_slots)
        # assign expect workers
        cur_assigns = {op.key: self._to_band(op.expect_worker)
                       for op in start_ops if op.expect_worker is not None}
        logger.debug('Start to assign %s start chunks for task %s',
                     len(start_ops), self._task.task_id)
        chunk_to_bands = assigner.assign(cur_assigns=cur_assigns)
        logger.debug('Assigned %s start chunks for task %s',
                     len(start_ops), self._task.task_id)

        # fuse node
        if self._fuse_enabled:
            logger.debug('Start to fuse chunks for task %s', self._task.task_id)
            chunk_to_bands = self._fuse(chunk_to_bands)
            logger.debug('Chunks fused for task %s', self._task.task_id)

        subtask_graph = SubtaskGraph()
        chunk_to_priorities = dict()
        chunk_to_fetch_chunk = dict()
        chunk_to_subtask = dict()
        visited = set()
        for chunk in self._chunk_graph.topological_iter():
            if chunk in visited:
                continue
            if isinstance(chunk.op, Fetch):
                continue

            # gen subtask
            subtask = self._gen_subtask(
                chunk, chunk_to_priorities,
                chunk_to_bands, chunk_to_fetch_chunk)

            for out in chunk.op.outputs:
                chunk_to_subtask[out] = subtask
                visited.add(out)
            subtask_graph.add_node(subtask)
            inp_subtasks = \
                [chunk_to_subtask[inp_chunk] for inp_chunk
                 in self._chunk_graph.predecessors(chunk)
                 if not isinstance(inp_chunk.op, Fetch)]
            for inp_subtask in inp_subtasks:
                subtask_graph.add_edge(inp_subtask, subtask)

        return subtask_graph



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
from typing import Dict, List

from .... import oscar as mo
from ....core import ChunkGraph
from ....core.operand import Fuse
from ....optimization.logical import OptimizationRecords
from ....typing import BandType, TileableType
from ....utils import get_params_fields
from ...scheduling import SchedulingAPI
from ...subtask import Subtask, SubtaskGraph, SubtaskResult, SubtaskStatus
from ...meta import MetaAPI
from ..core import Task, TaskResult, TaskStatus

logger = logging.getLogger(__name__)


class TaskStageProcessor:
    def __init__(self,
                 stage_id: str,
                 task: Task,
                 chunk_graph: ChunkGraph,
                 subtask_graph: SubtaskGraph,
                 bands: List[BandType],
                 tileable_to_subtasks: Dict[TileableType, List[Subtask]],
                 tileable_id_to_tileable: Dict[str, TileableType],
                 optimization_records: OptimizationRecords,
                 scheduling_api: SchedulingAPI,
                 meta_api: MetaAPI):
        self.stage_id = stage_id
        self.task = task
        self.chunk_graph = chunk_graph
        self.subtask_graph = subtask_graph
        self.tileable_to_subtasks = tileable_to_subtasks
        self.tileable_id_to_tileable = tileable_id_to_tileable
        self._bands = bands
        self._optimization_records = optimization_records

        # APIs
        self._scheduling_api = scheduling_api
        self._meta_api = meta_api

        # gen subtask_id to subtask
        self.subtask_id_to_subtask = {subtask.subtask_id: subtask
                                      for subtask in subtask_graph}
        self._subtask_to_bands: Dict[Subtask, BandType] = dict()
        self.subtask_snapshots: Dict[Subtask, SubtaskResult] = dict()
        self.subtask_results: Dict[Subtask, SubtaskResult] = dict()
        self._submitted_subtask_ids = set()

        self._band_manager: Dict[BandType, mo.ActorRef] = dict()

        # result
        self.result = TaskResult(
            task.task_id, task.session_id, self.stage_id,
            status=TaskStatus.pending, start_time=time.time())
        # status
        self._done = asyncio.Event()
        self._cancelled = asyncio.Event()

    def is_cancelled(self):
        return self._cancelled.is_set()

    async def _schedule_subtasks(self, subtasks: List[Subtask]):
        if not subtasks:
            return
        self._submitted_subtask_ids.update(subtask.subtask_id for subtask in subtasks)
        return await self._scheduling_api.add_subtasks(
            subtasks, [subtask.priority for subtask in subtasks])

    async def _update_chunks_meta(self, chunk_graph: ChunkGraph):
        get_meta = []
        chunks = chunk_graph.result_chunks
        for chunk in chunks:
            if isinstance(chunk.op, Fuse):
                chunk = chunk.chunk
            fields = get_params_fields(chunk)
            get_meta.append(self._meta_api.get_chunk_meta.delay(
                chunk.key, fields=fields))
        metas = await self._meta_api.get_chunk_meta.batch(*get_meta)
        for chunk, meta in zip(chunks, metas):
            chunk.params = meta
            original_chunk = \
                self._optimization_records.get_original_chunk(chunk)
            if original_chunk is not None:
                original_chunk.params = chunk.params

    def _schedule_done(self):
        self._done.set()

    async def set_subtask_result(self, result: SubtaskResult):
        subtask = self.subtask_id_to_subtask[result.subtask_id]
        self.subtask_results[subtask] = result.merge_bands(self.subtask_results.get(subtask))
        self._submitted_subtask_ids.difference_update([result.subtask_id])

        all_done = len(self.subtask_results) == len(self.subtask_graph)
        error_or_cancelled = result.status in (SubtaskStatus.errored, SubtaskStatus.cancelled)

        if all_done or error_or_cancelled:
            # terminated
            if all_done and not error_or_cancelled:
                # subtask graph finished, update result chunks' meta
                await self._update_chunks_meta(self.chunk_graph)

            # tell scheduling to finish subtasks
            await self._scheduling_api.finish_subtasks([result.subtask_id],
                                                       schedule_next=not error_or_cancelled)
            if self.result.status != TaskStatus.terminated:
                self.result = TaskResult(
                    self.task.task_id, self.task.session_id,
                    self.stage_id, start_time=self.result.start_time,
                    end_time=time.time(), status=TaskStatus.terminated,
                    error=result.error, traceback=result.traceback)
                if not all_done and error_or_cancelled:
                    if result.status == SubtaskStatus.errored:
                        logger.exception('Subtask %s errored', subtask.subtask_id,
                                         exc_info=(type(result.error), result.error, result.traceback))
                    # if error or cancel, cancel all submitted subtasks
                    await self._scheduling_api.cancel_subtasks(
                        list(self._submitted_subtask_ids))
                self._schedule_done()
        else:
            # not terminated, push success subtasks to queue if they are ready
            to_schedule_subtasks = []
            for succ_subtask in self.subtask_graph.successors(subtask):
                if succ_subtask in self.subtask_results:  # pragma: no cover
                    continue
                pred_subtasks = self.subtask_graph.predecessors(succ_subtask)
                if all(pred_subtask in self.subtask_results
                       for pred_subtask in pred_subtasks):
                    # all predecessors finished
                    to_schedule_subtasks.append(succ_subtask)
            await self._schedule_subtasks(to_schedule_subtasks)
            await self._scheduling_api.finish_subtasks([result.subtask_id])

    async def run(self):
        if len(self.subtask_graph) == 0:
            # no subtask to schedule, set status to done
            self._schedule_done()
            self.result.status = TaskStatus.terminated
            return

        # schedule independent subtasks
        indep_subtasks = list(self.subtask_graph.iter_indep())
        await self._schedule_subtasks(indep_subtasks)

        # wait for completion
        await self._done.wait()

    async def cancel(self):
        if self._done.is_set():  # pragma: no cover
            # already finished, ignore cancel
            return
        self._cancelled.set()
        # cancel running subtasks
        await self._scheduling_api.cancel_subtasks(
            list(self._submitted_subtask_ids))
        self._done.set()

    def error_or_cancelled(self) -> bool:
        if self.result.error is not None:
            return True
        if self.is_cancelled():
            return True
        return False



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ... import tensor as mt
from ...core import ENTITY_TYPE, recursive_tile
from ...core.context import get_context
from ...serialization.serializables import AnyField, BoolField, KeyField
from ...tensor.core import TensorOrder
from ..operands import LearnOperand, LearnOperandMixin, OutputType
from ..preprocessing import LabelBinarizer
from ..utils import check_array, check_consistent_length
from ._check_targets import _check_targets


class AccuracyScore(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.ACCURACY_SCORE

    _y_true = AnyField('y_true')
    _y_pred = AnyField('y_pred')
    _normalize = BoolField('normalize')
    _sample_weight = AnyField('sample_weight')
    _type_true = KeyField('type_true')

    def __init__(self, y_true=None, y_pred=None, normalize=None,
                 sample_weight=None, type_true=None, **kw):
        super().__init__(_y_true=y_true, _y_pred=y_pred,
                         _normalize=normalize, _sample_weight=sample_weight,
                         _type_true=type_true, **kw)
        self.output_types = [OutputType.tensor]

    @property
    def y_true(self):
        return self._y_true

    @property
    def y_pred(self):
        return self._y_pred

    @property
    def normalize(self):
        return self._normalize

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def type_true(self):
        return self._type_true

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        if self._y_true is not None:
            self._y_true = next(inputs_iter)
        if self._y_pred is not None:
            self._y_pred = next(inputs_iter)
        if self._type_true is not None:
            self._type_true = next(inputs_iter)
        if isinstance(self._sample_weight, ENTITY_TYPE):
            self._sample_weight = next(inputs_iter)

    def __call__(self, y_true, y_pred):
        type_true, y_true, y_pred = _check_targets(y_true, y_pred)
        self._type_true = type_true
        inputs = [y_true, y_pred, type_true]
        if isinstance(self._sample_weight, ENTITY_TYPE):
            inputs.append(self._sample_weight)

        dtype = np.dtype(float) if self._normalize else np.result_type(y_true, y_pred)
        return self.new_tileable(inputs, dtype=dtype,
                                 shape=(), order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        # make sure type_true executed first
        chunks = [op.type_true.chunks[0]]
        yield chunks

        ctx = get_context()
        type_true = ctx.get_chunks_result([chunks[0].key])[0]

        y_true, y_pred = op.y_true, op.y_pred
        if type_true.item().startswith('multilabel'):
            differing_labels = mt.count_nonzero(y_true - y_pred, axis=1)
            score = mt.equal(differing_labels, 0)
        else:
            score = mt.equal(y_true, y_pred)

        result = _weighted_sum(score, op.sample_weight, op.normalize)
        return [(yield from recursive_tile(result))]


def _weighted_sum(sample_score, sample_weight, normalize=False):
    if normalize:
        return mt.average(sample_score, weights=sample_weight)
    elif sample_weight is not None:
        return mt.dot(sample_score, sample_weight)
    else:
        return sample_score.sum()


def accuracy_score(y_true, y_pred, normalize=True, sample_weight=None,
                   session=None, run_kwargs=None):
    """Accuracy classification score.

    In multilabel classification, this function computes subset accuracy:
    the set of labels predicted for a sample must *exactly* match the
    corresponding set of labels in y_true.

    Read more in the :ref:`User Guide <accuracy_score>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator tensor / sparse tensor
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator tensor / sparse tensor
        Predicted labels, as returned by a classifier.

    normalize : bool, optional (default=True)
        If ``False``, return the number of correctly classified samples.
        Otherwise, return the fraction of correctly classified samples.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    Returns
    -------
    score : float
        If ``normalize == True``, return the fraction of correctly
        classified samples (float), else returns the number of correctly
        classified samples (int).

        The best performance is 1 with ``normalize == True`` and the number
        of samples with ``normalize == False``.

    See also
    --------
    jaccard_score, hamming_loss, zero_one_loss

    Notes
    -----
    In binary and multiclass classification, this function is equal
    to the ``jaccard_score`` function.

    Examples
    --------
    >>> from mars.learn.metrics import accuracy_score
    >>> y_pred = [0, 2, 1, 3]
    >>> y_true = [0, 1, 2, 3]
    >>> accuracy_score(y_true, y_pred).execute()
    0.5
    >>> accuracy_score(y_true, y_pred, normalize=False).execute()
    2

    In the multilabel case with binary label indicators:

    >>> import mars.tensor as mt
    >>> accuracy_score(mt.array([[0, 1], [1, 1]]), mt.ones((2, 2))).execute()
    0.5
    """

    # Compute accuracy for each possible representation
    op = AccuracyScore(y_true=y_true, y_pred=y_pred, normalize=normalize,
                       sample_weight=sample_weight)
    score = op(y_true, y_pred)
    return score.execute(session=session, **(run_kwargs or dict()))


def log_loss(y_true, y_pred, *, eps=1e-15, normalize=True, sample_weight=None,
             labels=None):
    r"""Log loss, aka logistic loss or cross-entropy loss.

    This is the loss function used in (multinomial) logistic regression
    and extensions of it such as neural networks, defined as the negative
    log-likelihood of a logistic model that returns ``y_pred`` probabilities
    for its training data ``y_true``.
    The log loss is only defined for two or more labels.
    For a single sample with true label :math:`y \in \{0,1\}` and
    and a probability estimate :math:`p = \operatorname{Pr}(y = 1)`, the log
    loss is:

    .. math::
        L_{\log}(y, p) = -(y \log (p) + (1 - y) \log (1 - p))

    Read more in the :ref:`User Guide <log_loss>`.

    Parameters
    ----------
    y_true : array-like or label indicator matrix
        Ground truth (correct) labels for n_samples samples.

    y_pred : array-like of float, shape = (n_samples, n_classes) or (n_samples,)
        Predicted probabilities, as returned by a classifier's
        predict_proba method. If ``y_pred.shape = (n_samples,)``
        the probabilities provided are assumed to be that of the
        positive class. The labels in ``y_pred`` are assumed to be
        ordered alphabetically, as done by
        :class:`preprocessing.LabelBinarizer`.

    eps : float, default=1e-15
        Log loss is undefined for p=0 or p=1, so probabilities are
        clipped to max(eps, min(1 - eps, p)).

    normalize : bool, default=True
        If true, return the mean loss per sample.
        Otherwise, return the sum of the per-sample losses.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    labels : array-like, default=None
        If not provided, labels will be inferred from y_true. If ``labels``
        is ``None`` and ``y_pred`` has shape (n_samples,) the labels are
        assumed to be binary and are inferred from ``y_true``.

    Returns
    -------
    loss : float

    Notes
    -----
    The logarithm used is the natural logarithm (base-e).

    Examples
    --------
    >>> from mars.learn.metrics import log_loss
    >>> log_loss(["spam", "ham", "ham", "spam"],
    ...          [[.1, .9], [.9, .1], [.8, .2], [.35, .65]])
    0.21616...

    References
    ----------
    C.M. Bishop (2006). Pattern Recognition and Machine Learning. Springer,
    p. 209.
    """
    y_pred = check_array(y_pred, ensure_2d=False)
    check_consistent_length(y_pred, y_true, sample_weight)

    lb = LabelBinarizer()

    if labels is not None:
        lb.fit(labels)
    else:
        lb.fit(y_true)

    if len(lb.classes_) == 1:
        if labels is None:
            raise ValueError('y_true contains only one label ({0}). Please '
                             'provide the true labels explicitly through the '
                             'labels argument.'.format(lb.classes_[0].fetch()))
        else:
            raise ValueError('The labels array needs to contain at least two '
                             'labels for log_loss, '
                             'got {0}.'.format(lb.classes_.fetch()))

    transformed_labels = lb.transform(y_true)

    if transformed_labels.shape[1] == 1:
        transformed_labels = mt.append(1 - transformed_labels,
                                       transformed_labels, axis=1)

    # Clipping
    y_pred = mt.clip(y_pred, eps, 1 - eps)

    # If y_pred is of single dimension, assume y_true to be binary
    # and then check.
    if y_pred.ndim == 1:  # pragma: no cover
        y_pred = y_pred[:, mt.newaxis]
    if y_pred.shape[1] == 1:  # pragma: no cover
        y_pred = mt.append(1 - y_pred, y_pred, axis=1)

    # Check if dimensions are consistent.
    transformed_labels = check_array(transformed_labels)
    if len(lb.classes_) != y_pred.shape[1]:
        if labels is None:
            raise ValueError("y_true and y_pred contain different number of "
                             "classes {0}, {1}. Please provide the true "
                             "labels explicitly through the labels argument. "
                             "Classes found in "
                             "y_true: {2}".format(transformed_labels.shape[1],
                                                  y_pred.shape[1],
                                                  lb.classes_.fetch()))
        else:
            raise ValueError('The number of classes in labels is different '
                             'from that in y_pred. Classes found in '
                             'labels: {0}'.format(lb.classes_.fetch()))

    # Renormalize
    y_pred /= y_pred.sum(axis=1)[:, mt.newaxis]
    loss = -(transformed_labels * mt.log(y_pred)).sum(axis=1)

    return _weighted_sum(loss, sample_weight, normalize).execute()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ..core import ENTITY_TYPE
from ..serialization.serializables import SerializableMeta
from ..tensor import tensor as astensor, stack
from ..tensor.array_utils import is_cupy
from ..tensor.core import TENSOR_TYPE
from ..utils import ceildiv, lazy_import
from .core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE, DataFrame as _Frame, \
    Series as _Series, Index as _Index
from .datasource.dataframe import from_pandas as from_pandas_df
from .datasource.series import from_pandas as from_pandas_series
from .datasource.index import from_pandas as from_pandas_index, \
    from_tileable as from_tileable_index
from .datasource.from_tensor import dataframe_from_tensor, series_from_tensor, \
    dataframe_from_1d_tileables
from .utils import is_index, is_cudf

cudf = lazy_import('cudf', globals=globals())


class InitializerMeta(SerializableMeta):
    def __instancecheck__(cls, instance):
        return isinstance(instance, (cls.__base__,) + getattr(cls, '_allow_data_type_'))


class DataFrame(_Frame, metaclass=InitializerMeta):
    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False,
                 chunk_size=None, gpu=None, sparse=None, num_partitions=None):
        need_repart = False
        if isinstance(data, TENSOR_TYPE):
            if chunk_size is not None:
                data = data.rechunk(chunk_size)
            df = dataframe_from_tensor(data, index=index, columns=columns, gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        elif isinstance(data, SERIES_TYPE):
            df = data.to_frame()
            need_repart = num_partitions is not None
        elif isinstance(data, DATAFRAME_TYPE):
            if not hasattr(data, 'data'):
                # DataFrameData
                df = _Frame(data)
            else:
                df = data
            need_repart = num_partitions is not None
        elif isinstance(data, dict) and any(isinstance(v, ENTITY_TYPE) for v in data.values()):
            # data is a dict and some value is tensor
            df = dataframe_from_1d_tileables(
                data, index=index, columns=columns, gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        elif isinstance(data, list) and any(isinstance(v, ENTITY_TYPE) for v in data):
            # stack data together
            data = stack(data)
            df = dataframe_from_tensor(data, index=index, columns=columns,
                                       gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        elif isinstance(index, (INDEX_TYPE, SERIES_TYPE)):
            if isinstance(data, dict):
                data = {k: astensor(v, chunk_size=chunk_size) for k, v in data.items()}
                df = dataframe_from_1d_tileables(data, index=index, columns=columns,
                                                 gpu=gpu, sparse=sparse)
            else:
                if data is not None:
                    data = astensor(data, chunk_size=chunk_size)
                df = dataframe_from_tensor(data, index=index,
                                           columns=columns, gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        else:
            if is_cudf(data) or is_cupy(data):  # pragma: no cover
                pdf = cudf.DataFrame(data, index=index, columns=columns, dtype=dtype)
                if copy:
                    pdf = pdf.copy()
            else:
                pdf = pd.DataFrame(data, index=index, columns=columns, dtype=dtype, copy=copy)
            if num_partitions is not None:
                chunk_size = ceildiv(len(pdf), num_partitions)
            df = from_pandas_df(pdf, chunk_size=chunk_size, gpu=gpu, sparse=sparse)

        if need_repart:
            df = df.rebalance(num_partitions=num_partitions)
        super().__init__(df.data)


class Series(_Series, metaclass=InitializerMeta):
    def __init__(self, data=None, index=None, dtype=None, name=None, copy=False,
                 chunk_size=None, gpu=None, sparse=None, num_partitions=None):
        if dtype is not None:
            dtype = np.dtype(dtype)

        need_repart = False
        if isinstance(data, (TENSOR_TYPE, INDEX_TYPE)):
            if chunk_size is not None:
                data = data.rechunk(chunk_size)
            name = name or getattr(data, 'name', None)
            series = series_from_tensor(data, index=index, name=name, gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        elif isinstance(index, INDEX_TYPE):
            if data is not None:
                data = astensor(data, chunk_size=chunk_size)
            series = series_from_tensor(data, index=index, name=name,
                                        dtype=dtype, gpu=gpu, sparse=sparse)
            need_repart = num_partitions is not None
        elif isinstance(data, SERIES_TYPE):
            if not hasattr(data, 'data'):
                # SeriesData
                series = _Series(data)
            else:
                series = data
            need_repart = num_partitions is not None
        else:
            if is_cudf(data) or is_cupy(data):  # pragma: no cover
                pd_series = cudf.Series(data, index=index, dtype=dtype, name=name)
                if copy:
                    pd_series = pd_series.copy()
            else:
                pd_series = pd.Series(data, index=index, dtype=dtype, name=name, copy=copy)
            if num_partitions is not None:
                chunk_size = ceildiv(len(pd_series), num_partitions)
            series = from_pandas_series(pd_series, chunk_size=chunk_size, gpu=gpu, sparse=sparse)

        if need_repart:
            series = series.rebalance(num_partitions=num_partitions)
        super().__init__(series.data)


class Index(_Index, metaclass=InitializerMeta):
    def __new__(cls, data, **_):
        # just return cls always until we support other Index's initializers
        return object.__new__(cls)

    def __init__(self, data=None, dtype=None, copy=False, name=None, tupleize_cols=True,
                 chunk_size=None, gpu=None, sparse=None, names=None, num_partitions=None,
                 store_data=False):
        need_repart = False
        if isinstance(data, INDEX_TYPE):
            if not hasattr(data, 'data'):
                # IndexData
                index = _Index(data)
            else:
                index = data
            need_repart = num_partitions is not None
        else:
            if isinstance(data, ENTITY_TYPE):
                name = name if name is not None else getattr(data, 'name', None)
                index = from_tileable_index(data, dtype=dtype, name=name, names=names)
                need_repart = num_partitions is not None
            else:
                if not is_index(data):
                    name = name if name is not None else getattr(data, 'name', None)
                    xdf = cudf if is_cudf(data) or is_cupy(data) else pd
                    try:
                        pd_index = xdf.Index(data=data, dtype=dtype, copy=copy, name=name,
                                             tupleize_cols=tupleize_cols)
                    except TypeError:  # pragma: no cover
                        pd_index = xdf.Index(data=data, dtype=dtype, copy=copy, name=name)
                else:
                    pd_index = data

                if num_partitions is not None:
                    chunk_size = ceildiv(len(pd_index), num_partitions)
                index = from_pandas_index(pd_index, chunk_size=chunk_size, gpu=gpu,
                                          sparse=sparse, store_data=store_data)

        if need_repart:
            index = index.rebalance(num_partitions=num_partitions)
        super().__init__(index.data)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Dict

from ..serialization.serializables import Serializable, StringField
from ..serialization.serializables.core import SerializableSerializer
from ..utils import tokenize


class Base(Serializable):
    _no_copy_attrs_ = {'_id'}
    _init_update_key_ = True

    _key = StringField('key')
    _id = StringField('id')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self._init_update_key_ and (not hasattr(self, '_key') or not self._key):
            self._update_key()
        if not hasattr(self, '_id') or not self._id:
            self._id = str(id(self))

    @property
    def _keys_(self):
        cls = type(self)
        member = '__keys_' + cls.__name__
        try:
            return getattr(cls, member)
        except AttributeError:
            slots = sorted(self._FIELDS)
            setattr(cls, member, slots)
            return slots

    @property
    def _values_(self):
        return [getattr(self, k, None) for k in self._keys_
                if k not in self._no_copy_attrs_]

    def __mars_tokenize__(self):
        if hasattr(self, '_key'):
            return self._key
        else:
            return (type(self), *self._values_)

    def _obj_set(self, k, v):
        object.__setattr__(self, k, v)

    def _update_key(self):
        self._obj_set('_key', tokenize(type(self).__name__, *self._values_))
        return self

    def reset_key(self):
        self._obj_set('_key', None)
        return self

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(_key=self.key))

    def copy_to(self, target):
        for attr in self._FIELDS:
            if (attr.startswith('__') and attr.endswith('__')) or attr in self._no_copy_attrs_:
                # we don't copy id to identify that the copied one is new
                continue
            try:
                attr_val = getattr(self, attr)
            except AttributeError:
                continue
            setattr(target, attr, attr_val)

        return target

    def copy_from(self, obj):
        obj.copy_to(self)

    @property
    def key(self):
        return self._key

    @property
    def id(self):
        return self._id


def buffered(func):
    @wraps(func)
    def wrapped(self, obj: Base, context: Dict):
        obj_id = (obj.key, obj.id)
        if obj_id in context:
            return {
                       'id': id(context[obj_id]),
                       'serializer': 'ref',
                       'buf_num': 0,
                   }, []
        else:
            context[obj_id] = obj
            return func(self, obj, context)
    return wrapped


class BaseSerializer(SerializableSerializer):
    @buffered
    def serialize(self, obj: Serializable, context: Dict):
        return (yield from super().serialize(obj, context))


BaseSerializer.register(Base)


class MarsError(Exception):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from dataclasses import asdict
from typing import Dict, List, Set

from .... import oscar as mo
from ....utils import implements
from ....typing import BandType
from ..core import _CommonMeta, _ChunkMeta
from .base import AbstractMetaStore, register_meta_store


@register_meta_store
class DictMetaStore(AbstractMetaStore):
    name = 'dict'

    def __init__(self, session_id: str, **kw):
        super().__init__(session_id)
        self._store: Dict[str, _CommonMeta] = dict()
        self._band_chunks: Dict[BandType, Set[str]] = defaultdict(set)
        if kw:  # pragma: no cover
            raise TypeError(f'Keyword arguments {kw!r} cannot be recognized.')

    @classmethod
    @implements(AbstractMetaStore.create)
    async def create(cls, config) -> Dict:
        # Nothing needs to do for dict-based meta store.
        # no extra kwargs.
        return dict()

    def _set_meta(self,
                  object_id: str,
                  meta: _CommonMeta):
        self._store[object_id] = meta
        if isinstance(meta, _ChunkMeta):
            for band in meta.bands:
                self._band_chunks[band].add(object_id)

    @implements(AbstractMetaStore.set_meta)
    @mo.extensible
    async def set_meta(self,
                       object_id: str,
                       meta: _CommonMeta):
        self._set_meta(object_id, meta)

    @set_meta.batch
    async def batch_set_meta(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._set_meta(*args, **kwargs)

    def _get_meta(self,
                  object_id: str,
                  fields: List[str] = None,
                  error: str = 'raise') -> Dict:
        if error not in ('raise', 'ignore'):  # pragma: no cover
            raise ValueError('error must be raise or ignore')
        try:
            meta = asdict(self._store[object_id])
            if fields:
                return {k: meta[k] for k in fields}
            return meta
        except KeyError:
            if error == 'raise':
                raise
            else:
                return

    @implements(AbstractMetaStore.get_meta)
    @mo.extensible
    async def get_meta(self,
                       object_id: str,
                       fields: List[str] = None,
                       error: str = 'raise') -> Dict:
        return self._get_meta(object_id, fields=fields, error=error)

    @get_meta.batch
    async def batch_get_meta(self, args_list, kwargs_list):
        metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            metas.append(self._get_meta(*args, **kwargs))
        return metas

    def _del_meta(self, object_id: str):
        meta = self._store[object_id]
        if isinstance(meta, _ChunkMeta):
            for band in meta.bands:
                chunks = self._band_chunks[band]
                chunks.remove(object_id)
                if len(chunks) == 0:
                    del self._band_chunks[band]
        del self._store[object_id]

    @implements(AbstractMetaStore.del_meta)
    @mo.extensible
    async def del_meta(self,
                       object_id: str):
        self._del_meta(object_id)

    @del_meta.batch
    async def batch_del_meta(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._del_meta(*args, **kwargs)

    def _add_chunk_bands(self,
                         object_id: str,
                         bands: List[BandType]):
        meta = self._store[object_id]
        assert isinstance(meta, _ChunkMeta)
        meta.bands = list(set(meta.bands) | set(bands))
        for band in bands:
            self._band_chunks[band].add(object_id)

    @implements(AbstractMetaStore.add_chunk_bands)
    @mo.extensible
    async def add_chunk_bands(self,
                              object_id: str,
                              bands: List[BandType]):
        self._add_chunk_bands(object_id, bands)

    @add_chunk_bands.batch
    async def batch_add_chunk_bands(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._add_chunk_bands(*args, **kwargs)

    def _remove_chunk_bands(self,
                            object_id: str,
                            bands: List[BandType]):
        meta = self._store[object_id]
        assert isinstance(meta, _ChunkMeta)
        meta.bands = list(set(meta.bands) - set(bands))
        for band in bands:
            self._band_chunks[band].remove(object_id)

    @implements(AbstractMetaStore.remove_chunk_bands)
    @mo.extensible
    async def remove_chunk_bands(self,
                                 object_id: str,
                                 bands: List[BandType]):
        self._remove_chunk_bands(object_id, bands)

    @remove_chunk_bands.batch
    async def batch_remove_chunk_bands(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._remove_chunk_bands(*args, **kwargs)

    async def get_band_chunks(self, band: BandType) -> List[str]:
        return list(self._band_chunks[band])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections import defaultdict
from functools import lru_cache
from typing import List, Dict

from .. import oscar as mo
from ..lib.aio import new_isolation
from ..core.context import Context
from ..typing import BandType, SessionType
from ..storage.base import StorageLevel
from ..utils import implements
from .cluster import ClusterAPI, NodeRole
from .session import SessionAPI
from .storage import StorageAPI
from .subtask import SubtaskAPI
from .meta import MetaAPI


class ThreadedServiceContext(Context):
    _cluster_api: ClusterAPI
    _session_api: SessionAPI
    _meta_api: MetaAPI
    _subtask_api: SubtaskAPI

    def __init__(self,
                 session_id: str,
                 supervisor_address: str,
                 worker_address: str,
                 current_address: str,
                 loop: asyncio.AbstractEventLoop,
                 band: BandType = None):
        super().__init__(session_id=session_id,
                         supervisor_address=supervisor_address,
                         worker_address=worker_address,
                         current_address=current_address,
                         band=band)
        self._loop = loop
        # new isolation with current loop,
        # so that session created in tile and execute
        # can get the right isolation
        new_isolation(loop=self._loop, threaded=False)

        self._running_session_id = None
        self._running_op_key = None

        # APIs
        self._cluster_api = None
        self._session_api = None
        self._meta_api = None
        self._subtask_api = None

    async def init(self):
        self._cluster_api = await ClusterAPI.create(
            self.supervisor_address)
        self._session_api = await SessionAPI.create(
            self.supervisor_address)
        self._meta_api = await MetaAPI.create(
            self.session_id, self.supervisor_address)
        try:
            self._subtask_api = await SubtaskAPI.create(
                self.current_address)
        except mo.ActorNotExist:
            pass

    def _call(self, coro):
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return fut.result()

    @implements(Context.get_current_session)
    def get_current_session(self) -> SessionType:
        from ..deploy.oscar.session import new_session

        return new_session(self.supervisor_address, self.session_id, default=False)

    @implements(Context.get_supervisor_addresses)
    def get_supervisor_addresses(self) -> List[str]:
        return self._call(self._cluster_api.get_supervisors())

    @implements(Context.get_worker_addresses)
    def get_worker_addresses(self) -> List[str]:
        return list(self._call(
            self._cluster_api.get_nodes_info(role=NodeRole.WORKER)))

    @implements(Context.get_total_n_cpu)
    def get_total_n_cpu(self) -> int:
        all_bands = self._call(
            self._cluster_api.get_all_bands())
        n_cpu = 0
        for band, size in all_bands.items():
            _, band_name = band
            if band_name.startswith('numa-'):
                n_cpu += size
        return n_cpu

    async def _get_chunks_meta(self,
                               data_keys: List[str],
                               fields: List[str] = None,
                               error: str = 'raise') -> List[Dict]:
        # get chunks meta
        get_metas = []
        for data_key in data_keys:
            meta = self._meta_api.get_chunk_meta.delay(
                data_key, fields=fields, error=error)
            get_metas.append(meta)
        metas = await self._meta_api.get_chunk_meta.batch(*get_metas)
        return metas

    async def _get_chunks_result(self, data_keys: List[str]) -> List:
        metas = await self._get_chunks_meta(data_keys, fields=['bands'])
        addresses = [meta['bands'][0][0] for meta in metas]

        storage_api_to_gets = defaultdict(lambda: (list(), list()))
        for data_key, address in zip(data_keys, addresses):
            storage_api = await StorageAPI.create(self.session_id, address)
            storage_api_to_gets[storage_api][0].append(data_key)
            storage_api_to_gets[storage_api][1].append(
                storage_api.get.delay(data_key))
        results = dict()
        for storage_api, (keys, gets) in storage_api_to_gets.items():
            chunks_data = await storage_api.get.batch(*gets)
            for chunk_key, chunk_data in zip(keys, chunks_data):
                results[chunk_key] = chunk_data
        return [results[key] for key in data_keys]

    @implements(Context.get_chunks_result)
    def get_chunks_result(self,
                          data_keys: List[str]) -> List:
        return self._call(self._get_chunks_result(data_keys))

    @implements(Context.get_chunks_meta)
    def get_chunks_meta(self,
                        data_keys: List[str],
                        fields: List[str] = None,
                        error='raise') -> List[Dict]:
        return self._call(self._get_chunks_meta(data_keys, fields=fields, error=error))

    async def _get_backend_info(self,
                                address: str = None,
                                level: StorageLevel = StorageLevel.MEMORY) -> dict:
        if address is None:
            address = self.worker_address
        storage_api = await StorageAPI.create(self.session_id, address)
        return await storage_api.get_storage_info(level)

    @implements(Context.get_storage_info)
    def get_storage_info(self,
                         address: str = None,
                         level: StorageLevel = StorageLevel.MEMORY):
        return self._call(self._get_backend_info(address, level))

    @implements(Context.create_remote_object)
    def create_remote_object(self,
                             name: str,
                             object_cls, *args, **kwargs):
        ref = self._call(self._session_api.create_remote_object(
            self.session_id, name, object_cls, *args, **kwargs))
        return _RemoteObjectWrapper(ref, self._loop)

    @implements(Context.get_remote_object)
    def get_remote_object(self, name: str):
        ref = self._call(self._session_api.get_remote_object(
            self.session_id, name))
        return _RemoteObjectWrapper(ref, self._loop)

    @implements(Context.destroy_remote_object)
    def destroy_remote_object(self,
                              name: str):
        return self._call(self._session_api.destroy_remote_object(
            self.session_id, name))

    @implements(Context.register_custom_log_path)
    def register_custom_log_path(self,
                                 session_id: str,
                                 tileable_op_key: str,
                                 chunk_op_key: str,
                                 worker_address: str,
                                 log_path: str):
        return self._call(self._session_api.register_custom_log_path(
            session_id, tileable_op_key, chunk_op_key, worker_address, log_path))

    @implements(Context.new_custom_log_dir)
    @lru_cache(50)
    def new_custom_log_dir(self) -> str:
        return self._call(self._session_api.new_custom_log_dir(
            self.current_address, self.session_id))

    def set_running_operand_key(self, session_id: str, op_key: str):
        self._running_session_id = session_id
        self._running_op_key = op_key

    def set_progress(self, progress: float):
        if self._running_op_key is None or self._subtask_api is None:  # pragma: no cover
            return
        return self._call(self._subtask_api.set_running_operand_progress(
            session_id=self._running_session_id, op_key=self._running_op_key,
            slot_address=self.current_address, progress=progress))


class _RemoteObjectWrapper:
    def __init__(self,
                 ref: mo.ActorRef,
                 loop: asyncio.AbstractEventLoop):
        self._ref = ref
        self._loop = loop

    def __getattr__(self, attr):
        func = getattr(self._ref, attr)

        def wrap(*args, **kwargs):
            coro = func(*args, **kwargs)
            fut = asyncio.run_coroutine_threadsafe(coro, loop=self._loop)
            return fut.result()

        return wrap



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from .core import AbstractMetaAPI


class MetaWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = '/api/session/(?P<session_id>[^/]+)/meta'

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_meta_api(self, session_id: str):
        from .oscar import MetaAPI
        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys([session_id])
        return await MetaAPI.create(session_id, address)

    @web_api('(?P<data_key>[^/]+)', method='get')
    async def get_chunk_meta(self, session_id: str, data_key: str):
        fields_str = self.get_argument('fields', None)
        error = self.get_argument('error', 'raise')
        fields = fields_str.split(',') if fields_str else None

        oscar_api = await self._get_oscar_meta_api(session_id)
        result = await oscar_api.get_chunk_meta(data_key, fields=fields, error=error)
        self.write(serialize_serializable(result))


web_handlers = {
    MetaWebAPIHandler.get_root_pattern(): MetaWebAPIHandler
}


class WebMetaAPI(AbstractMetaAPI, MarsWebAPIClientMixin):
    def __init__(self, session_id: str, address: str):
        self._session_id = session_id
        self._address = address.rstrip('/')

    @mo.extensible
    async def get_chunk_meta(self,
                             object_id: str,
                             fields: List[str] = None,
                             error: str = 'raise') -> Optional[Dict]:
        params = dict(error=error)
        req_addr = f'{self._address}/api/session/{self._session_id}/meta/{object_id}'
        if fields:
            params['fields'] = ','.join(fields)
        res = await self._request_url('GET', req_addr, params=params)
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...config import options
from ...core import OutputType, ENTITY_TYPE
from ...serialization.serializables import BoolField
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameKurtosis(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = opcodes.KURTOSIS
    _func_name = 'kurt'

    _bias = BoolField('bias')
    _fisher = BoolField('fisher')

    def __init__(self, bias=None, fisher=None, **kw):
        super().__init__(_bias=bias, _fisher=fisher, **kw)

    @property
    def bias(self):
        return self._bias

    @property
    def fisher(self):
        return self._fisher

    @classmethod
    def get_reduction_callable(cls, op):
        from .aggregation import where_function
        skipna, bias, fisher = op.skipna, op.bias, op.fisher

        def kurt(x):
            cnt = x.count()
            mean = x.mean(skipna=skipna)
            divided = (x ** 4).mean(skipna=skipna) \
                - 4 * (x ** 3).mean(skipna=skipna) * mean \
                + 6 * (x ** 2).mean(skipna=skipna) * mean ** 2 \
                - 3 * mean ** 4
            var = x.var(skipna=skipna, ddof=0)
            if isinstance(var, ENTITY_TYPE) or var > 0:
                val = where_function(var > 0, divided / var ** 2, np.nan)
            else:
                val = np.nan
            if not bias:
                val = where_function((var > 0) & (cnt > 3),
                                     (val * (cnt ** 2 - 1) - 3 * (cnt - 1) ** 2) / (cnt - 2) / (cnt - 3),
                                     np.nan)
            if not fisher:
                val += 3
            return val

        return kurt


def kurt_series(df, axis=None, skipna=None, level=None, combine_size=None, bias=False,
                fisher=True, method=None):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameKurtosis(axis=axis, skipna=skipna, level=level, combine_size=combine_size,
                           bias=bias, fisher=fisher, output_types=[OutputType.scalar],
                           use_inf_as_na=use_inf_as_na, method=method)
    return op(df)


def kurt_dataframe(df, axis=None, skipna=None, level=None, numeric_only=None, combine_size=None,
                   bias=False, fisher=True, method=None):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameKurtosis(axis=axis, skipna=skipna, level=level, numeric_only=numeric_only,
                           bias=bias, fisher=fisher, combine_size=combine_size,
                           output_types=[OutputType.series], use_inf_as_na=use_inf_as_na,
                           method=method)
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..proxima.simple_index import build_index, search_index


METRIC_TO_PROXIMA_METRIC_TYPE = {
    'l2': 'Euclidean',
    'euclidean': 'Euclidean',
    'canberra': 'Canberra',
    'chebyshev': 'Chebyshev',
    'sqeuclidean': 'SquaredEuclidean',
    'innerproduct': 'InnerProduct',
    'manhattan': 'Manhattan'
}

build_proxima_index = build_index
proxima_query = search_index



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import FieldTypes, KeyField, \
    TupleField, StringField
from ...core import ExecutableTuple
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..core import TensorOrder


class TensorUnravelIndex(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.UNRAVEL_INDEX

    _input = KeyField('input')
    _dims = TupleField('dims', FieldTypes.int32)
    _order = StringField('order')

    def __init__(self, dims=None, order=None, **kw):
        super().__init__(_dims=dims, _order=order, **kw)
        if self._order is None:
            self._order = 'C'

    @property
    def dims(self):
        return self._dims

    @property
    def order(self):
        return self._order

    @property
    def output_limit(self):
        return float('inf')

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, indices):
        order = TensorOrder.C_ORDER if self._order == 'C' else TensorOrder.F_ORDER
        kws = [{'pos': i, 'order': order} for i in range(len(self._dims))]
        return ExecutableTuple(self.new_tensors([indices], indices.shape, kws=kws, output_limit=len(kws)))

    @classmethod
    def tile(cls, op):
        indices = op.inputs[0]
        dims = op.dims
        order = op.outputs[0].order

        out_chunks = [list() for _ in range(len(dims))]
        for in_chunk in indices.chunks:
            chunk_op = op.copy().reset_key()
            chunk_kws = [{'pos': i, 'index': in_chunk.index, 'order': order}
                         for i in range(len(dims))]
            chunks = chunk_op.new_chunks([in_chunk], shape=in_chunk.shape, kws=chunk_kws,
                                         output_limit=len(dims))
            for out_chunk, c in zip(out_chunks, chunks):
                out_chunk.append(c)

        new_op = op.copy()
        kws = [{'chunks': out_chunk, 'nsplits': indices.nsplits, 'shape': o.shape}
               for out_chunk, o in zip(out_chunks, op.outputs)]
        return new_op.new_tensors(op.inputs, kws=kws, output_limit=len(dims), order=order)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)
        indices = inputs[0]

        with device(device_id):
            outputs = xp.unravel_index(indices, op.dims, order=op.order)
            for o, output in zip(op.outputs, outputs):
                ctx[o.key] = output


def unravel_index(indices, dims, order='C'):
    """
    Converts a flat index or tensor of flat indices into a tuple
    of coordinate tensors.

    Parameters
    ----------
    indices : array_like
        An integer tensor whose elements are indices into the flattened
        version of a tensor of dimensions ``dims``.
    dims : tuple of ints
        The shape of the tensor to use for unraveling ``indices``.
    order : {'C', 'F'}, optional
        Determines whether the indices should be viewed as indexing in
        row-major (C-style) or column-major (Fortran-style) order.

    Returns
    -------
    unraveled_coords : tuple of Tensor
        Each tensor in the tuple has the same shape as the ``indices``
        tensor.

    See Also
    --------
    ravel_multi_index

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.unravel_index([22, 41, 37], (7,6)).execute()
    (array([3, 6, 6]), array([4, 5, 1]))

    >>> mt.unravel_index(1621, (6,7,8,9)).execute()
    (3, 1, 4, 1)
    """
    indices = astensor(indices)
    if isinstance(dims, Iterable):
        dims = tuple(dims)
    else:
        dims = (dims,)

    if order not in 'CF':
        raise TypeError("only 'C' or 'F' order is permitted")

    op = TensorUnravelIndex(dims=dims, dtype=np.dtype(np.intp), order=order)
    return op(indices)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from typing import List, Dict, Optional, Set, Type, TypeVar

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....typing import BandType
from ...core import NodeRole
from ..core import watch_method, NodeStatus, WorkerSlotInfo, QuotaInfo, \
    DiskInfo, StorageInfo
from .core import AbstractClusterAPI

APIType = TypeVar('APIType', bound='ClusterAPI')
logger = logging.getLogger(__name__)


class ClusterAPI(AbstractClusterAPI):
    def __init__(self, address: str):
        self._address = address
        self._locator_ref = None
        self._uploader_ref = None

    async def _init(self):
        from ..locator import SupervisorLocatorActor
        from ..uploader import NodeInfoUploaderActor

        self._locator_ref = await mo.actor_ref(SupervisorLocatorActor.default_uid(),
                                               address=self._address)
        self._uploader_ref = await mo.actor_ref(NodeInfoUploaderActor.default_uid(),
                                                address=self._address)

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls: Type[APIType], address: str) -> APIType:
        api_obj = cls(address)
        await api_obj._init()
        return api_obj

    @alru_cache(cache_exceptions=False)
    async def _get_node_info_ref(self):
        from ..supervisor.node_info import NodeInfoCollectorActor
        [node_info_ref] = await self.get_supervisor_refs(
            [NodeInfoCollectorActor.default_uid()])
        return node_info_ref

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        return await self._locator_ref.get_supervisors(filter_ready=filter_ready)

    @watch_method
    async def watch_supervisors(self,
                                version: Optional[int] = None):
        return await self._locator_ref.watch_supervisors(version=version)

    async def get_supervisors_by_keys(self, keys: List[str]) -> List[str]:
        """
        Get supervisor address hosting the specified key

        Parameters
        ----------
        keys
            key for a supervisor address
        watch
            if True, will watch changes of supervisor changes

        Returns
        -------
        out
            addresses of the supervisor
        """
        get_supervisor = self._locator_ref.get_supervisor
        return await get_supervisor.batch(
            *(get_supervisor.delay(k) for k in keys)
        )

    @watch_method
    async def watch_supervisors_by_keys(self, keys: List[str],
                                        version: Optional[int] = None):
        return await self._locator_ref.watch_supervisors_by_keys(keys, version=version)

    async def get_supervisor_refs(self, uids: List[str]) -> List[mo.ActorRef]:
        """
        Get actor references hosting the specified actor uid

        Parameters
        ----------
        uids
            uids for a supervisor address
        watch
            if True, will watch changes of supervisor changes

        Returns
        -------
        out : List[mo.ActorRef]
            references of the actors
        """
        addrs = await self.get_supervisors_by_keys(uids)
        return await asyncio.gather(*[
            mo.actor_ref(uid, address=addr) for addr, uid in zip(addrs, uids)
        ])

    async def watch_supervisor_refs(self, uids: List[str]):
        async for addrs in self.watch_supervisors_by_keys(uids):
            yield await asyncio.gather(*[
                mo.actor_ref(uid, address=addr) for addr, uid in zip(addrs, uids)
            ])

    @watch_method
    async def watch_nodes(self, role: NodeRole, env: bool = False,
                          resource: bool = False, detail: bool = False,
                          version: Optional[int] = None,
                          statuses: Set[NodeStatus] = None,
                          exclude_statuses: Set[NodeStatus] = None) -> List[Dict[str, Dict]]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.watch_nodes(
            role, env=env, resource=resource, detail=detail,
            statuses=statuses, version=version)

    async def get_nodes_info(self, nodes: List[str] = None, role: NodeRole = None,
                             env: bool = False, resource: bool = False, detail: bool = False,
                             statuses: Set[NodeStatus] = None,
                             exclude_statuses: Set[NodeStatus] = None):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_nodes_info(
            nodes=nodes, role=role, env=env, resource=resource,
            detail=detail, statuses=statuses)

    async def set_node_status(self, node: str, role: NodeRole, status: NodeStatus):
        """
        Set status of node

        Parameters
        ----------
        node : str
            address of node
        role: NodeRole
            role of node
        status : NodeStatus
            status of node
        """
        node_info_ref = await self._get_node_info_ref()
        await node_info_ref.update_node_info(node, role, status=status)

    async def get_all_bands(self, role: NodeRole = None,
                            statuses: Set[NodeStatus] = None,
                            exclude_statuses: Set[NodeStatus] = None) -> Dict[BandType, int]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_all_bands(role, statuses=statuses)

    @watch_method
    async def watch_all_bands(self, role: NodeRole = None,
                              version: Optional[int] = None,
                              statuses: Set[NodeStatus] = None,
                              exclude_statuses: Set[NodeStatus] = None):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.watch_all_bands(
            role, statuses=statuses, version=version)

    async def get_mars_versions(self) -> List[str]:
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_mars_versions()

    async def get_bands(self) -> Dict:
        """
        Get bands that can be used for computation on current node.

        Returns
        -------
        band_to_slots : dict
            Band to n_slot.
        """
        return await self._uploader_ref.get_bands()

    async def mark_node_ready(self):
        """
        Mark current node ready for work loads
        """
        await self._uploader_ref.mark_node_ready()

    async def wait_node_ready(self):
        """
        Wait current node to be ready
        """
        await self._uploader_ref.wait_node_ready()

    async def wait_all_supervisors_ready(self):
        """
        Wait till all expected supervisors are ready
        """
        await self._locator_ref.wait_all_supervisors_ready()

    async def set_band_slot_infos(self, band_name: str,
                                  slot_infos: List[WorkerSlotInfo]):
        await self._uploader_ref.set_band_slot_infos.tell(band_name, slot_infos)

    async def set_band_quota_info(self, band_name: str,
                                  quota_info: QuotaInfo):
        await self._uploader_ref.set_band_quota_info.tell(band_name, quota_info)

    async def set_node_disk_info(self, disk_info: List[DiskInfo]):
        await self._uploader_ref.set_node_disk_info(disk_info)

    @mo.extensible
    async def set_band_storage_info(self, band_name: str, storage_info: StorageInfo):
        await self._uploader_ref.set_band_storage_info(band_name, storage_info)

    async def request_worker(
            self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None) -> str:
        node_allocator_ref = await self._get_node_allocator_ref()
        address = await node_allocator_ref.request_worker(worker_cpu, worker_mem, timeout)
        return address

    async def release_worker(self, address: str):
        node_allocator_ref = await self._get_node_allocator_ref()
        await node_allocator_ref.release_worker(address)
        node_info_ref = await self._get_node_info_ref()
        await node_info_ref.update_node_info(address, NodeRole.WORKER, status=NodeStatus.STOPPED)

    async def reconstruct_worker(self, address: str):
        node_allocator_ref = await self._get_node_allocator_ref()
        await node_allocator_ref.reconstruct_worker(address)

    @alru_cache(cache_exceptions=False)
    async def _get_node_allocator_ref(self):
        from ..supervisor.node_allocator import NodeAllocatorActor
        [node_allocator_ref] = await self.get_supervisor_refs(
            [NodeAllocatorActor.default_uid()])
        return node_allocator_ref


class MockClusterAPI(ClusterAPI):
    @classmethod
    async def create(cls: Type[APIType], address: str, **kw) -> APIType:
        from ..supervisor.locator import SupervisorPeerLocatorActor
        from ..uploader import NodeInfoUploaderActor
        from ..supervisor.node_allocator import NodeAllocatorActor
        from ..supervisor.node_info import NodeInfoCollectorActor

        dones, _ = await asyncio.wait([
            mo.create_actor(SupervisorPeerLocatorActor, 'fixed', address,
                            uid=SupervisorPeerLocatorActor.default_uid(),
                            address=address),
            mo.create_actor(NodeInfoCollectorActor,
                            uid=NodeInfoCollectorActor.default_uid(),
                            address=address),
            mo.create_actor(NodeAllocatorActor, 'fixed', address,
                            uid=NodeAllocatorActor.default_uid(),
                            address=address),
            mo.create_actor(NodeInfoUploaderActor, NodeRole.WORKER,
                            interval=kw.get('upload_interval'),
                            band_to_slots=kw.get('band_to_slots'),
                            use_gpu=kw.get('use_gpu', False),
                            uid=NodeInfoUploaderActor.default_uid(),
                            address=address),
        ])

        for task in dones:
            try:
                task.result()
            except mo.ActorAlreadyExist:  # pragma: no cover
                pass

        api = await super().create(address=address)
        await api.mark_node_ready()
        return api

    @classmethod
    async def cleanup(cls, address: str):
        from ..supervisor.locator import SupervisorPeerLocatorActor
        from ..uploader import NodeInfoUploaderActor
        from ..supervisor.node_info import NodeInfoCollectorActor

        await asyncio.wait([
            mo.destroy_actor(mo.create_actor_ref(
                uid=SupervisorPeerLocatorActor.default_uid(),
                address=address)),
            mo.destroy_actor(mo.create_actor_ref(
                uid=NodeInfoCollectorActor.default_uid(),
                address=address)),
            mo.destroy_actor(mo.create_actor_ref(
                uid=NodeInfoUploaderActor.default_uid(),
                address=address)),
        ])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ....utils import lazy_import
from ...backend import BaseActorBackend, register_backend
from ..context import MarsActorContext
from .driver import RayActorDriver
from .pool import RayMainPool
from .utils import process_address_to_placement, get_placement_group

ray = lazy_import("ray")

__all__ = ['RayActorBackend']


@register_backend
class RayActorBackend(BaseActorBackend):
    @staticmethod
    def name():
        return "ray"

    @staticmethod
    def get_context_cls():
        return MarsActorContext

    @staticmethod
    def get_driver_cls():
        return RayActorDriver

    @staticmethod
    async def create_actor_pool(
        address: str,
        n_process: int = None,
        **kwargs
    ):
        # pop `n_io_process` from kwargs as ray doesn't need this
        kwargs.pop('n_io_process', 0)
        pg_name, bundle_index, _ = process_address_to_placement(address)
        pg = get_placement_group(pg_name) if pg_name else None
        if not pg:
            bundle_index = -1
        num_cpus = kwargs.get('main_pool_cpus', 0)
        actor_handle = ray.remote(RayMainPool).options(
            num_cpus=num_cpus, name=address,
            max_concurrency=10000,  # By default, 1000 tasks can be running concurrently.
            max_restarts=-1,  # Auto restarts by ray
            placement_group=pg, placement_group_bundle_index=bundle_index).remote(
                address, n_process, **kwargs)
        await actor_handle.start.remote()
        return actor_handle



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....storage import StorageLevel
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import DataInfo
from .core import AbstractStorageAPI


class StorageWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = '/api/session/(?P<session_id>[^/]+)/storage'

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_meta_api(self, session_id: str):
        from ...meta import MetaAPI
        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys([session_id])
        return await MetaAPI.create(session_id, address)

    async def _get_storage_api_by_object_id(self, session_id: str,
                                            object_id: str):
        from .oscar import StorageAPI
        meta_api = await self._get_oscar_meta_api(session_id)
        bands = (await meta_api.get_chunk_meta(object_id, ['bands'])).get('bands')
        if not bands:
            raise KeyError
        return await StorageAPI.create(session_id, bands[0][0], bands[0][1])

    @web_api('(?P<data_key>[^/]+)', method='get')
    async def get_data(self, session_id: str, data_key: str):
        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        result = await oscar_api.get(data_key)
        self.write(serialize_serializable(result))

    @web_api('(?P<data_key>[^/]+)', method='post')
    async def get_data_by_post(self, session_id: str, data_key: str):
        body_args = deserialize_serializable(self.request.body) if self.request.body else None
        conditions = body_args.get('conditions')

        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        result = await oscar_api.get(data_key, conditions)
        self.write(serialize_serializable(result))

    @web_api('(?P<data_key>[^/]+)', method='put')
    async def put_data(self, session_id: str, data_key: str):
        level = self.get_argument('level', None) or 'MEMORY'
        level = getattr(StorageLevel, level.upper())

        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        res = await oscar_api.put(
            data_key, deserialize_serializable(self.request.body), level)
        self.write(serialize_serializable(res))


web_handlers = {
    StorageWebAPIHandler.get_root_pattern(): StorageWebAPIHandler
}


class WebStorageAPI(AbstractStorageAPI, MarsWebAPIClientMixin):
    def __init__(self,
                 session_id: str,
                 address: str,
                 band_name: str):
        self._session_id = session_id
        self._address = address.rstrip('/')
        self._band_name = band_name

    @mo.extensible
    async def get(self,
                  data_key: str,
                  conditions: List = None,
                  error: str = 'raise') -> Any:
        path = f'{self._address}/api/session/{self._session_id}/storage/{data_key}'
        params = dict(error=error)
        if conditions is not None:
            params['conditions'] = conditions
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path, method='POST',
            headers={'Content-Type': 'application/octet-stream'},
            data=body)
        return deserialize_serializable(res.body)

    async def put(self, data_key: str,
                  obj: object,
                  level: StorageLevel = StorageLevel.MEMORY) -> DataInfo:
        params = dict(level=level.name.lower())
        path = f'{self._address}/api/session/{self._session_id}/storage/{data_key}'
        res = await self._request_url(
            path=path, method='PUT', params=params,
            headers={'Content-Type': 'application/octet-stream'},
            data=serialize_serializable(obj))
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, List, Optional, Set

from ....lib.aio import alru_cache
from ....typing import BandType
from ....utils import serialize_serializable, deserialize_serializable
from ...core import NodeRole
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import watch_method, NodeStatus
from .core import AbstractClusterAPI


class ClusterWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = '/api/cluster'

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @staticmethod
    def _convert_node_dict(node_info_list: Dict[str, Dict]):
        res = {}
        for node_addr, node in node_info_list.items():
            res_dict = node.copy()
            res_dict['status'] = res_dict['status'].value
            res[node_addr] = res_dict
        return res

    @web_api('nodes', method=['get', 'post'])
    async def get_nodes_info(self):
        watch = bool(int(self.get_argument('watch', '0')))
        env = bool(int(self.get_argument('env', '0')))
        resource = bool(int(self.get_argument('resource', '0')))
        detail = bool(int(self.get_argument('detail', '0')))

        nodes_arg = self.get_argument('nodes', None)
        nodes = nodes_arg.split(',') if nodes_arg is not None else None

        role_arg = self.get_argument('role', None)
        role = NodeRole(int(role_arg)) if role_arg is not None else None

        statuses_arg = self.get_argument('statuses', None)
        statuses = set(NodeStatus(int(v)) for v in statuses_arg.split(',')) \
            if statuses_arg else None

        exclude_statuses_arg = self.get_argument('exclude_statuses', None)
        exclude_statuses = set(NodeStatus(int(v)) for v in exclude_statuses_arg.split(',')) \
            if exclude_statuses_arg else None

        statuses = WebClusterAPI._calc_statuses(statuses, exclude_statuses)

        cluster_api = await self._get_cluster_api()
        result = {}
        if watch:
            assert nodes is None
            version = self.get_argument('version', '') or None
            if version:
                version = int(version)

            async for version, node_infos in cluster_api.watch_nodes(
                role, env=env, resource=resource, detail=detail,
                statuses=statuses, version=version
            ):
                result['version'] = version
                result['nodes'] = self._convert_node_dict(node_infos)
                break
        else:
            nodes = await cluster_api.get_nodes_info(
                nodes=nodes, role=role, env=env, resource=resource,
                statuses=statuses, detail=detail
            )
            result['nodes'] = self._convert_node_dict(nodes)
        self.write(json.dumps(result))

    @web_api('bands', method='get')
    async def get_all_bands(self):
        role_arg = self.get_argument('role', None)
        role = NodeRole(int(role_arg)) if role_arg is not None else None
        watch = bool(int(self.get_argument('watch', '0')))

        statuses_arg = self.get_argument('statuses', None)
        statuses = set(NodeStatus(int(v)) for v in statuses_arg.split(',')) \
            if statuses_arg else None

        cluster_api = await self._get_cluster_api()
        if watch:
            version = self.get_argument('version', '') or None
            if version:
                version = int(version)

            async for version, bands in cluster_api.watch_all_bands(
                role, statuses=statuses, version=version
            ):
                self.write(serialize_serializable((version, bands)))
                break
        else:
            self.write(serialize_serializable(
                await cluster_api.get_all_bands(role, statuses=statuses)
            ))

    @web_api('versions', method='get')
    async def get_mars_versions(self):
        cluster_api = await self._get_cluster_api()
        self.write(json.dumps(list(await cluster_api.get_mars_versions())))


web_handlers = {
    ClusterWebAPIHandler.get_root_pattern(): ClusterWebAPIHandler
}


class WebClusterAPI(AbstractClusterAPI, MarsWebAPIClientMixin):
    def __init__(self, address: str):
        self._address = address.rstrip('/')

    @staticmethod
    def _convert_node_dict(node_info_list: Dict[str, Dict]):
        res = {}
        for node_addr, node in node_info_list.items():
            res_dict = node.copy()
            res_dict['status'] = NodeStatus(res_dict['status'])
            res[node_addr] = res_dict
        return res

    async def _get_nodes_info(self, nodes: List[str] = None, role: NodeRole = None,
                              env: bool = False, resource: bool = False, detail: bool = False,
                              watch: bool = False, statuses: Set[NodeStatus] = None,
                              version: Optional[int] = None):
        statuses_str = ','.join(str(status.value) for status in statuses) if statuses else ''
        args = [
            ('nodes', ','.join(nodes) if nodes else None),
            ('role', role.value if role is not None else None),
            ('env', 1 if env else 0),
            ('resource', 1 if resource else 0),
            ('detail', 1 if detail else 0),
            ('watch', 1 if watch else 0),
            ('statuses', statuses_str),
            ('version', str(version or '')),
        ]
        args_str = '&'.join(f'{key}={val}' for key, val in args if val is not None)

        path = f'{self._address}/api/cluster/nodes'
        res = await self._request_url(
            path=path, method='POST', data=args_str,
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
        )
        result = json.loads(res.body)
        if watch:
            return result['version'], self._convert_node_dict(result['nodes'])
        else:
            return self._convert_node_dict(result['nodes'])

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        statuses = {NodeStatus.READY} if filter_ready \
            else {NodeStatus.STARTING, NodeStatus.READY}
        res = await self._get_nodes_info(role=NodeRole.SUPERVISOR, statuses=statuses)
        return list(res.keys())

    @watch_method
    async def watch_supervisors(self, version: Optional[int] = None):
        version, res = await self._get_nodes_info(role=NodeRole.SUPERVISOR,
                                                  watch=True, version=version)
        return version, list(res.keys())

    async def get_nodes_info(self, nodes: List[str] = None,
                             role: NodeRole = None,
                             env: bool = False,
                             resource: bool = False,
                             detail: bool = False,
                             statuses: Set[NodeStatus] = None,
                             exclude_statuses: Set[NodeStatus] = None):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        return await self._get_nodes_info(nodes, role=role, env=env, resource=resource,
                                          detail=detail, watch=False, statuses=statuses)

    @watch_method
    async def watch_nodes(self, role: NodeRole, env: bool = False,
                          resource: bool = False, detail: bool = False,
                          statuses: Set[NodeStatus] = None,
                          exclude_statuses: Set[NodeStatus] = None,
                          version: Optional[int] = None) -> List[Dict[str, Dict]]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        return await self._get_nodes_info(role=role, env=env, resource=resource,
                                          detail=detail, watch=True, statuses=statuses,
                                          version=version)

    async def get_all_bands(self, role: NodeRole = None,
                            statuses: Set[NodeStatus] = None,
                            exclude_statuses: Set[NodeStatus] = None) -> Dict[BandType, int]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        statuses_str = ','.join(str(status.value) for status in statuses) if statuses else ''
        params = {}
        if role is not None:  # pragma: no cover
            params['role'] = role.value
        if statuses_str:
            params['statuses'] = statuses_str

        path = f'{self._address}/api/cluster/bands'
        res = await self._request_url('GET', path, params=params)
        return deserialize_serializable(res.body)

    @watch_method
    async def watch_all_bands(self, role: NodeRole = None,
                              statuses: List[NodeStatus] = None,
                              exclude_statuses: Set[NodeStatus] = None,
                              version: Optional[int] = None):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        statuses_str = ','.join(str(status.value) for status in statuses) if statuses else ''
        params = dict(watch=1, version=str(version or ''))
        if role is not None:  # pragma: no cover
            params['role'] = role.value
        if statuses_str:
            params['statuses'] = statuses_str

        path = f'{self._address}/api/cluster/bands'
        res = await self._request_url('GET', path, params=params)
        return deserialize_serializable(res.body)

    async def get_mars_versions(self) -> List[str]:
        path = f'{self._address}/api/cluster/versions'
        res = await self._request_url('GET', path)
        return list(json.loads(res.body))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import uuid
from typing import List, Optional, Union

from ...core import recursive_tile
from ...core.context import get_context, Context
from ...oscar import ActorNotExist
from ...serialization.serializables import Int64Field, StringField
from ...typing import TileableType, OperandType
from ..core import IndexValue
from ..operands import DataFrameOperand, DataFrameOperandMixin


class HeadOptimizedDataSource(DataFrameOperand, DataFrameOperandMixin):
    __slots__ = ()
    # Data source op that optimized for head,
    # First, it will try to trigger first_chunk.head() and raise TilesError,
    # When iterative tiling is triggered,
    # check if the first_chunk.head() meets requirements.
    _nrows = Int64Field('nrows')

    @property
    def nrows(self):
        return self._nrows

    @property
    def first_chunk(self):
        return getattr(self, '_first_chunk', None)

    @classmethod
    def _tile(cls, op):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _tile_head(cls, op: "HeadOptimizedDataSource"):
        tileds = cls._tile(op)
        chunks = tileds[0].chunks

        # execute first chunk
        yield chunks[:1]

        ctx = get_context()
        chunk_shape = ctx.get_chunks_meta([chunks[0].key])[0]['shape']

        if chunk_shape[0] == op.nrows:
            # the first chunk has enough data
            tileds[0]._nsplits = tuple((s,) for s in chunk_shape)
            chunks[0]._shape = chunk_shape
            tileds[0]._chunks = chunks[:1]
            tileds[0]._shape = chunk_shape
        else:
            for chunk in tileds[0].chunks:
                chunk.op._nrows = None
            # otherwise
            tiled = yield from recursive_tile(tileds[0].iloc[:op.nrows])
            tileds = [tiled]
        return tileds

    @classmethod
    def tile(cls, op: "HeadOptimizedDataSource"):
        if op.nrows is not None:
            return (yield from cls._tile_head(op))
        else:
            return cls._tile(op)


class ColumnPruneSupportedDataSourceMixin(DataFrameOperandMixin):
    __slots__ = ()

    def get_columns(self):  # pragma: no cover
        raise NotImplementedError

    def set_pruned_columns(self, columns, *, keep_order=None):  # pragma: no cover
        raise NotImplementedError


class _IncrementalIndexRecorder:
    _done: List[Optional[asyncio.Event]]
    _chunk_sizes: List[Optional[int]]

    def __init__(self, n_chunk: int):
        self._n_chunk = n_chunk
        self._done = [asyncio.Event() for _ in range(n_chunk)]
        self._chunk_sizes = [None] * n_chunk
        self._waiters = set()

    def _can_destroy(self):
        return all(e.is_set() for e in self._done) and not self._waiters

    def add_waiter(self, i: int):
        self._waiters.add(i)

    async def wait(self, i: int):
        if i == 0:
            return 0, self._can_destroy()
        self._waiters.add(i)
        try:
            await asyncio.gather(*(e.wait() for e in self._done[:i]))
        finally:
            self._waiters.remove(i)
        # all chunk finished and no waiters
        return sum(self._chunk_sizes[:i]), self._can_destroy()

    async def finish(self, i: int, size: int):
        self._chunk_sizes[i] = size
        self._done[i].set()


class IncrementalIndexDatasource(HeadOptimizedDataSource):
    __slots__ = ()

    incremental_index_recorder_name = StringField('incremental_index_recorder_name')


class IncrementalIndexDataSourceMixin(DataFrameOperandMixin):
    __slots__ = ()

    @classmethod
    def post_tile(cls, op: OperandType, results: List[TileableType]):
        if op.incremental_index and results is not None and \
                isinstance(results[0].index_value.value, IndexValue.RangeIndex):
            result = results[0]
            for chunk in result.chunks:
                chunk.op.priority = -chunk.index[0]
            n_chunk = len(result.chunks)
            ctx = get_context()
            if ctx:
                name = str(uuid.uuid4())
                ctx.create_remote_object(
                    name, _IncrementalIndexRecorder, n_chunk)
                for chunk in result.chunks:
                    chunk.op.incremental_index_recorder_name = name

    @classmethod
    def pre_execute(cls, ctx: Union[dict, Context], op: OperandType):
        out = op.outputs[0]
        if op.incremental_index and \
                isinstance(out.index_value.value, IndexValue.RangeIndex) and \
                getattr(op, 'incremental_index_recorder_name', None):
            index = out.index[0]
            recorder_name = op.incremental_index_recorder_name
            recorder = ctx.get_remote_object(recorder_name)
            recorder.add_waiter(index)

    @classmethod
    def post_execute(cls, ctx: Union[dict, Context], op: OperandType):
        out = op.outputs[0]
        result = ctx[out.key]
        if op.incremental_index and \
                isinstance(out.index_value.value, IndexValue.RangeIndex) and \
                getattr(op, 'incremental_index_recorder_name', None):
            recorder_name = op.incremental_index_recorder_name
            recorder = ctx.get_remote_object(recorder_name)
            index = out.index[0]
            recorder.finish(index, len(result))
            # wait for previous chunks to finish, then update index
            size, can_destroy = recorder.wait(index)
            result.index += size
            if can_destroy:
                try:
                    ctx.destroy_remote_object(recorder_name)
                except ActorNotExist:
                    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ...resource import cpu_count, cuda_count
from ...services import NodeRole
from ...utils import get_next_port
from .cmdline import OscarCommandRunner
from .local import start_worker, stop_worker
from .pool import create_worker_actor_pool


class WorkerCommandRunner(OscarCommandRunner):
    command_description = 'Mars Worker'
    node_role = NodeRole.WORKER

    def __init__(self):
        super().__init__()
        self.band_to_slot = dict()
        self.cuda_devices = []
        self.n_io_process = 1

    def config_args(self, parser):
        super().config_args(parser)
        parser.add_argument('--n-cpu', help='num of CPU to use', default='auto')
        parser.add_argument('--n-io-process', help='num of IO processes', default='1')
        parser.add_argument('--cuda-devices',
                            help='CUDA device to use, if not specified, will use '
                                 'all available devices',
                            default='auto')

    def parse_args(self, parser, argv, environ=None):
        environ = environ or os.environ
        args = super().parse_args(parser, argv, environ=environ)

        if self.config.get('cluster', {}).get('backend', 'fixed') == 'fixed' \
                and not args.supervisors:  # pragma: no cover
            raise ValueError('--supervisors is needed to start Mars Worker')

        if args.endpoint is None:
            args.endpoint = f'{args.host}:{get_next_port()}'
        self.n_io_process = int(args.n_io_process)

        n_cpu = cpu_count() if args.n_cpu == 'auto' else int(args.n_cpu)

        if 'CUDA_VISIBLE_DEVICES' in os.environ:  # pragma: no cover
            args.cuda_devices = os.environ['CUDA_VISIBLE_DEVICES'].strip()

        if args.cuda_devices == 'auto':
            self.cuda_devices = list(range(cuda_count()))
        elif args.cuda_devices.strip() == '':  # pragma: no cover
            # allow using CPU only
            self.cuda_devices = []
        else:  # pragma: no cover
            self.cuda_devices = [int(i) for i in args.cuda_devices.split(',')]

        self.band_to_slot = band_to_slot = dict()
        band_to_slot['numa-0'] = n_cpu
        for i in self.cuda_devices:  # pragma: no cover
            band_to_slot[f'gpu-{i}'] = 1

        storage_config = self.config['storage'] = self.config.get('storage', {})
        backends = storage_config['backends'] = storage_config.get('backends', [])
        plasma_config = storage_config['plasma'] = storage_config.get('plasma', {})
        disk_config = storage_config['disk'] = storage_config.get('disk', {})
        if 'MARS_CACHE_MEM_SIZE' in environ:
            plasma_config['store_memory'] = environ['MARS_CACHE_MEM_SIZE']
        if 'MARS_PLASMA_DIRS' in environ:
            plasma_config['plasma_directory'] = environ['MARS_PLASMA_DIRS']
        if 'MARS_SPILL_DIRS' in environ:
            backends.append('disk')
            disk_config['root_dirs'] = environ['MARS_SPILL_DIRS']

        return args

    async def create_actor_pool(self):
        return await create_worker_actor_pool(
            self.args.endpoint, self.band_to_slot, ports=self.ports,
            n_io_process=self.n_io_process, modules=list(self.args.load_modules),
            logging_conf=self.logging_conf,
            cuda_devices=self.cuda_devices,
            subprocess_start_method='forkserver' if os.name != 'nt' else 'spawn'
        )

    async def start_services(self):
        return await start_worker(
            self.pool.external_address, self.args.supervisors,
            self.band_to_slot, list(self.args.load_modules), self.config)

    async def stop_services(self):
        return await stop_worker(self.pool.external_address, self.config)


main = WorkerCommandRunner()

if __name__ == '__main__':  # pragma: no branch
    main()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....oscar.errors import MarsError
from ....typing import BandType
from ....utils import dataslots
from ...subtask import Subtask, SubtaskResult, SubtaskStatus
from ...task import TaskAPI
from ..utils import redirect_subtask_errors

logger = logging.getLogger(__name__)


# the default times to reschedule subtask.
DEFAULT_SUBTASK_MAX_RESCHEDULES = 0


@dataslots
@dataclass
class SubtaskScheduleInfo:
    subtask: Subtask
    band_futures: Dict[BandType, asyncio.Future] = field(default_factory=dict)
    max_reschedules: int = 0
    num_reschedules: int = 0


class SubtaskManagerActor(mo.Actor):
    _subtask_infos: Dict[str, SubtaskScheduleInfo]  # key is subtask id

    @classmethod
    def gen_uid(cls, session_id: str):
        return f'{session_id}_subtask_manager'

    def __init__(self, session_id: str, subtask_max_reschedules: int = DEFAULT_SUBTASK_MAX_RESCHEDULES):
        self._session_id = session_id
        self._subtask_infos = dict()
        self._subtask_max_reschedules = subtask_max_reschedules

        self._queueing_ref = None
        self._global_slot_ref = None

    async def __post_create__(self):
        from .queueing import SubtaskQueueingActor
        self._queueing_ref = await mo.actor_ref(
            SubtaskQueueingActor.gen_uid(self._session_id), address=self.address)
        from ..supervisor import GlobalSlotManagerActor
        self._global_slot_ref = await mo.actor_ref(
            GlobalSlotManagerActor.default_uid(), address=self.address)

    @alru_cache
    async def _get_task_api(self):
        return await TaskAPI.create(self._session_id, self.address)

    async def add_subtasks(self, subtasks: List[Subtask], priorities: List[Tuple]):
        async with redirect_subtask_errors(self, subtasks):
            for subtask in subtasks:
                # the extra_config may be None. the extra config overwrites the default value.
                subtask_max_reschedules = (subtask.extra_config.get('subtask_max_reschedules')
                                           if subtask.extra_config else None)
                if subtask_max_reschedules is None:
                    subtask_max_reschedules = self._subtask_max_reschedules
                self._subtask_infos[subtask.subtask_id] = \
                    SubtaskScheduleInfo(subtask, max_reschedules=subtask_max_reschedules)

            virtual_subtasks = [subtask for subtask in subtasks if subtask.virtual]
            for subtask in virtual_subtasks:
                task_api = await self._get_task_api()
                await task_api.set_subtask_result(SubtaskResult(
                    subtask_id=subtask.subtask_id,
                    session_id=subtask.session_id,
                    task_id=subtask.task_id,
                    progress=1.0,
                    status=SubtaskStatus.succeeded))

            await self._queueing_ref.add_subtasks(
                [subtask for subtask in subtasks if not subtask.virtual], priorities)
            await self._queueing_ref.submit_subtasks.tell()

    @alru_cache(maxsize=10000)
    async def _get_execution_ref(self, band: BandType):
        from ..worker.execution import SubtaskExecutionActor
        return await mo.actor_ref(SubtaskExecutionActor.default_uid(), address=band[0])

    async def finish_subtasks(self, subtask_ids: List[str], schedule_next: bool = True):
        band_tasks = defaultdict(lambda: 0)
        for subtask_id in subtask_ids:
            subtask_info = self._subtask_infos.pop(subtask_id, None)
            if schedule_next and subtask_info is not None:
                for band in subtask_info.band_futures.keys():
                    band_tasks[band] += 1

        if band_tasks:
            coros = []
            for band, subtask_count in band_tasks.items():
                coros.append(self._queueing_ref.submit_subtasks.tell(band, subtask_count))
            await asyncio.wait(coros)

    def _get_subtasks_by_ids(self, subtask_ids: List[str]) -> List[Optional[Subtask]]:
        subtasks = []
        for stid in subtask_ids:
            try:
                subtasks.append(self._subtask_infos[stid].subtask)
            except KeyError:
                subtasks.append(None)
        return subtasks

    async def submit_subtask_to_band(self, subtask_id: str, band: BandType):
        async with redirect_subtask_errors(self, self._get_subtasks_by_ids([subtask_id])):
            try:
                subtask_info = self._subtask_infos[subtask_id]
                execution_ref = await self._get_execution_ref(band)
                task = asyncio.create_task(execution_ref.run_subtask(
                    subtask_info.subtask, band[1], self.address))
                subtask_info.band_futures[band] = task
                result = yield task
                task_api = await self._get_task_api()
                await task_api.set_subtask_result(result)
            except (OSError, MarsError) as ex:
                # TODO: We should handle ServerClosed Error.
                if subtask_info.subtask.retryable and subtask_info.num_reschedules < subtask_info.max_reschedules:
                    logger.error('Reschedule subtask %s due to %s',
                                 subtask_info.subtask.subtask_id, ex)
                    subtask_info.num_reschedules += 1
                    await self._queueing_ref.add_subtasks(
                            [subtask_info.subtask], [subtask_info.subtask.priority or tuple()])
                    await self._queueing_ref.submit_subtasks.tell()
                else:
                    raise ex
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                if subtask_info.subtask.retryable and subtask_info.num_reschedules < subtask_info.max_reschedules:
                    logger.error('Failed to reschedule subtask %s, '
                                 'num_reschedules: %s, max_reschedules: %s, unhandled exception: %s',
                                 subtask_info.subtask.subtask_id,
                                 subtask_info.num_reschedules, subtask_info.max_reschedules, ex)
                raise ex
            finally:
                # make sure slot is released before marking tasks as finished
                await self._global_slot_ref.release_subtask_slots(
                            band, subtask_info.subtask.session_id, subtask_info.subtask.subtask_id)
                logger.debug('Slot released for band %s after subtask %s',
                             band, subtask_info.subtask.subtask_id)

    async def cancel_subtasks(self, subtask_ids: List[str],
                              kill_timeout: Union[float, int] = 5):
        queued_subtask_ids = []
        single_cancel_tasks = []

        task_api = await self._get_task_api()

        async def cancel_single_task(subtask, raw_tasks, cancel_tasks):
            if cancel_tasks:
                await asyncio.wait(cancel_tasks)
            if raw_tasks:
                dones, _ = await asyncio.wait(raw_tasks)
            else:
                dones = []
            if not dones or all(fut.cancelled() for fut in dones):
                await task_api.set_subtask_result(SubtaskResult(
                    subtask_id=subtask.subtask_id,
                    session_id=subtask.session_id,
                    task_id=subtask.task_id,
                    status=SubtaskStatus.cancelled))

        for subtask_id in subtask_ids:
            if subtask_id not in self._subtask_infos:
                # subtask may already finished or not submitted at all
                continue

            subtask_info = self._subtask_infos[subtask_id]
            raw_tasks_to_cancel = list(subtask_info.band_futures.values())

            if not subtask_info.band_futures:
                queued_subtask_ids.append(subtask_id)
                single_cancel_tasks.append(asyncio.create_task(
                    cancel_single_task(subtask_info.subtask,
                                       raw_tasks_to_cancel, [])
                ))
            else:
                cancel_tasks = []
                for band in subtask_info.band_futures.keys():
                    execution_ref = await self._get_execution_ref(band)
                    cancel_tasks.append(asyncio.create_task(
                        execution_ref.cancel_subtask(subtask_id, kill_timeout=kill_timeout)
                    ))
                single_cancel_tasks.append(asyncio.create_task(
                    cancel_single_task(subtask_info.subtask,
                                       raw_tasks_to_cancel,
                                       cancel_tasks)
                ))
        if queued_subtask_ids:
            await self._queueing_ref.remove_queued_subtasks(queued_subtask_ids)
        if single_cancel_tasks:
            yield asyncio.wait(single_cancel_tasks)

        for subtask_id in subtask_ids:
            self._subtask_infos.pop(subtask_id, None)
